#! python3
# -*- coding: ISO-8859-1 -*-
#-------------------------------------------------------------------------------
# $Id$
#
# Copyright 2016 Jean-Luc PLOIX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#-------------------------------------------------------------------------------

'''Metaphor program NN and GM
'''

import os
import sys
sys.path.append('..')

import time
import importlib
from six.moves.configparser import NoOptionError
from collections import OrderedDict  
from random import randint

from ....nntoolbox.modelutils import getModelDict
from ....nntoolbox.modelutils import getGmxFileAndSave
from ....nntoolbox.filetoolbox import adaptConfigStr2hidden, IndexFilename, \
    getDictFromGmx, setEnvironFromFile
from ....nntoolbox.excelutils import  finishExcel, getCustomDict
from ....nntoolbox.utils import Namespace, finished, hasOneOf, \
    strintlist, doNothing, float_format, IDformat, roundoff_format, \
    make_extern 
from ....nntoolbox import utils as nnu
from ....nntoolbox.cmd_line_dialogs import yesNoQuestion, QuestionInterrupt
from ....nntoolbox.utils import toPrint, recupenv
from ....nntoolbox.timetb import second2str
from ....nntoolbox.lineartoolbox import gramSchmidt, cosTheta  
from ....nntoolbox.datareader.modeldataframe import get_modelDataframe
from ....nntoolbox.datareader.excelsource import getFrameFromExcel  
from ....nntoolbox.configparsertoolbox import defaultedConfigParser

from ....version import __version__ as metaphorversion 
from ....monal.util.monaltoolbox import  debugOutput
from ....monal.util.utils import getuserbase, CCharOnlys
from ....monal.modelutils import ReadModule

from ..use.nn_use import usage, usageInterne
from ..printfile import printInfo


nn1serv = None
#demodebug = nn1serv.demodebug & nn1serv.DEMO_DEBUG_DEMO
monalversion = metaphorversion
nntoolboxversion = metaphorversion

import warnings
warnings.filterwarnings("ignore", category=UserWarning)

debugrun = 0
debug = os.environ.get('DEBUG', '')
debugrun = debug.lower() == 'run'

fullActionNames = {
    'test': 'Test',
    'loo': 'Leave-One-Out Cross Validation',
    'lootest': 'Leave-One-Out test', 
    'bootstrap': 'Bootstrap', 
    'bootstraptest': 'Bootstrap test'}

namesdict = {
    "libfilename": "",
    "resultsheetname":'Training sumary',
    "trainsheetname": 'Training results',
    "testsheetname": 'Test results',
    "tstshname": "",
    "lootestsheetname": 'LOO test results',
    "lootstshname": "",
    "bstestsheetname": 'Bootstrap test results',
    "trnshname": "",
    "lootrainsheetname": 'LOO CV results',
    "bstrainsheetname": 'Bootstrap training results',
    }
#     "sumarysheetname":'Training summary',

def mainAcvtionFromTest(testaction):
    res = ""
    if testaction.endswith('test'):
        res = testaction[:-4]
        if not res:
            return "train"
    return res

def startExtraAction(options, config, module=None, trainDataFrame=None,
        testDataFrame=None, apiTrainRes=None,
        tempres="", actiontype="test", excelwriter=None,
        currentTrainFrame=None, sheetname="", resfile="", 
        doprint=doNothing, mynames=None):
    trainortest = ''
    startwithtest = currentTrainFrame is None
    startwithlootest = not 'loo' in options.actionlist
    startwithbstest = not 'bootstrap' in options.actionlist
    if actiontype.endswith('test'):
        trainortest = 'test'
        trainortest ='train'
    m0 = "launch {0}".format(fullActionNames[actiontype])
    m1 = "launch {0} ?".format(fullActionNames[actiontype])
    m2 = "continue and launch {0} ?".format(fullActionNames[actiontype])
    message0 = m1 if options.notrain else m2
    message = m0 if options.verbose else ""
    valdef = int(nn1serv.dialogdefaults['action']['continue'])
    if not yesNoQuestion(message0, valdef, message, forcedefault=options.yes, 
                         doprint=doprint):
        doprint("aborted")
        finished(tempres)
        return
    if options.verbose > 0:
        doprint("Running {0} {1}".format(fullActionNames[actiontype], trainortest))  
    
    if  actiontype in ['loo', 'bootstrap']:       
        datacount, datasize = 0, 0
        dico = None
        gmxfile = ""
        root = options.root
        if module is None:
            module = nn1serv.LoadModuleFromCfg(config)
        elif config.has_section(root):
            gmxfile = config.get(root, "gmxfile")
            dico = getDictFromGmx(gmxfile)
            module.loadParameters(dico)
        if trainDataFrame is not None:
            datacount, datasize = \
                module.loadTrainingDataFromFrame(trainDataFrame)
        return module, datacount, datasize, dico, gmxfile, startwithbstest
    
    elif actiontype in ['lootest', 'bootstraptest']:
        ResultTrainColumn = 0
        IDList = None
        startwithtest = ((actiontype == 'lootest') and (startwithlootest)) or \
            ((actiontype == 'bootstraptest') and (startwithbstest))
        if startwithtest and (options.verbose >= 3):  # and not testindex:
            foldername = make_extern(options.savedir, options)
            if options.verbose > 0:
                doprint("working folder: %s\n"% (foldername))  #outputmarker
            currentTrainFrame = getFrameFromExcel(options.origin, sheetname, 
                reduce=True)
            if currentTrainFrame is None:
                mess = "cannot find {}|{}".format(options.origin, sheetname)
                raise Exception(mess)
            ResultTrainColumn = testDataFrame.shape[1]  #LOOResultTrainColumn
        #ResultTrainColumn = sum([len(val) for val in options.datafields]) + 1 
        else:
            sheetname = ""  

        if apiTrainRes is None:
            extratype = "loo" if actiontype == 'lootest' else "bs"
            apiTrainRes = nn1serv.apiTrainResults(options, resfile, 
                                        extratype=extratype, mynames=mynames) 
        currentTrainFrame = apiTrainRes[0]
        IDlist = apiTrainRes[3]
        root = apiTrainRes[4] 
        ResultTrainColumn = apiTrainRes[5] 
        gmxfile = apiTrainRes[6]
        
        return None, currentTrainFrame, sheetname, ResultTrainColumn, IDList, \
            startwithtest

def basicTrainTestAction(options,
                         trainDataFrame, testDataFrame,
                         doprint, resultFrameList, excelactionlist,
                         mynames):
    """Train and test basic action.
    seed and hidden variables are fixed
    """
    excelwriter = None
    resultFrame = None
    trainFrame = None
    testFrame = None
    looTrainFrame = None
    looTestFrame = None
    bsTrainFrame = None
    bsTestFrame = None
    LOOResultTrainColumn = 0
    testresultColumn = -1
    startsloowithtest = True
    startsbswithtest = True
    loowidth = 0
    traintype = ""
    starttimeloc = time.time()
    if hasattr(options, "configstrMem"):
        options.configstr = adaptConfigStr2hidden(options.configstrMem, 
                                                  options.hidden)
    trainfailed = False
    resfile = os.path.join(options.savedir, 'train.cfg')
    sqlFileName = ""
    sqlLOOFileName = ""
    sqlBSFileName = ""
    gmxfile = ""
    gmxfileLOO = ""
    gmxfileBS = ""
    module = None

    if 'train' in options.actionlist:
        if options.verbose > 0:
            doprint("Running Train")

        rest = nn1serv.trainAction(options,
                                  sourceframe=trainDataFrame,
                                  doprint=doprint,
                                  resultFrameList=resultFrameList)
        libfilename, resultFrame, cfgfile, gmxfile, dico, sqlFileName = rest
        confResult = defaultedConfigParser(resfile)
        if not confResult.has_section(options.root):
            confResult.add_section(options.root)
        confResult.set(options.root, "cfgfile", cfgfile)
        confResult.save()

        trainfailed = resultFrame is None
        if not trainfailed:
            if "ID" in resultFrame.columns:
                resultFrame.set_index("ID", inplace=True)
            fname = make_extern(gmxfile, options)
            if options.verbose > 0:
                doprint("Writing model file: {0}".format(fname))
                doprint('Train done')
#         else:
#             doprint("train failed")
    if trainfailed:
        doprint('train failed')
        return None, None  # basicTrainTestAction
    
    maxrec = options.multimaxrec[0]
    if resultFrame is not None:
        maxrec = min(maxrec, resultFrame.shape[0])  # ici 2019/11/15
    options.maxrec = maxrec
    if options.copytype and hasOneOf(options.actionlist, excelactionlist):
        res = nn1serv.apiCreateExcelAndResultFrame(options, resultFrame,
            excelwriter, resfile, doprint) 
        excelwriter, resfile, resultFrame, tempres = res
        confResult = defaultedConfigParser(resfile)
        countmax = options.maxrec
        if resultFrame is not None:
            countmax = min(countmax, resultFrame.shape[0])
        root = options.root
        cfgfile = "" 
        try:
            cfgfile = confResult.get(root, "cfgfile")
        except:
            for sect in confResult.sections():
                if confResult.has_option(sect, "cfgfile"):
                    cfgfile = confResult.get(sect, "cfgfile")
                    root = sect
                    break
        config = defaultedConfigParser(cfgfile)
        paramcount = config.getint(root, "paramcount")
#    if resultFrame is not None:
    nn1serv.saveFrameToExcel(options, resultFrame, excelwriter,
        mynames.resultsheetname, actionName='result')        

    if options.notrain:
        if hasattr(options, "modelfile") and options.modelfile:
            gmxfile = options.modelfile
        else:
            gmxfile = config.get(root, 'gmxfile')
        if hasattr(options, "modelloofile") and options.modelloofile:
            gmxfileLOO = options.modelloofile
        elif config.has_option(root, 'gmxfileloo'):
            gmxfileLOO = config.get(root, 'gmxfileloo')
        if hasattr(options, "modelbsfile") and options.modelbsfile:
            gmxfileBS = options.modelbsfile
        elif config.has_option(root, 'gmxfilebs'):
            gmxfileBS = config.get(root, 'gmxfilebs')
 
        
        try:
            mcount = config.getint(root, "mcount")
        except NoOptionError:
            mcount = -1
            
        pptyname = config.get(root, "propertyname")
        paramnames = config.get(root, "paramnames").split(";")
        outputnorm = config.get(root, "outputnorm").split(";")
        outputnorm = [float(val) for val in outputnorm]
        dataCount = config.getint(root, "datacount")
        options.seed = config.getint("training_%s"% pptyname, 'seed')
        if hasattr(options, "sqlFileName") and options.sqlFileName:
            sqlFileName = options.sqlFileName
        else:
            sqlFileName = config.get(root, "sqlfilename")
        gm0 = getModelDict(config, propertyName=pptyname, 
            parameterNames=paramnames, outputnorm=outputnorm, 
            paramCount=paramcount, dataCount=dataCount, 
            version=metaphorversion)
        if resultFrame is None:
            if os.path.exists(gmxfile):
                dico = getDictFromGmx(gmxfile)            
        else:
            if "ID" in resultFrame.columns:
                indexlist = resultFrame["ID"][:countmax]
            else:
                indexlist = resultFrame.index[:countmax]
            config.set(root, "mcount", "%d" % countmax)
            if not gmxfile or not os.path.exists(gmxfile):
                gmxfile, dico = getGmxFileAndSave(sqlFileName, gm0, 
                    config, indexlist, options, full=True, 
                    dosave=(mcount != countmax))
                config.set(root, "gmxfile", gmxfile)
            else:
                dico = getDictFromGmx(gmxfile)
    
    apiTrainRes = nn1serv.apiTrainResults(options, resfile, 
        gmxfile=gmxfile, trainDataFrame=trainDataFrame, mynames=mynames) 
    trainFrame, looTrainFrame, bsTrainFrame, IDlist, root, \
        trainresultColumn, gmxfile = apiTrainRes # 
#     module.__del__()
#     module = None
    if options.notrain and (trainFrame is not None) and (trainFrame.shape[0] > dataCount):
        lst = list(trainFrame.index)[dataCount:]
        trainFrame.drop(lst, inplace = True)
        
    nn1serv.saveFrameToExcel(options, trainFrame, excelwriter, 
        mynames.trainsheetname, actionName='train')
    # if looTrainFrame is not None:
    nn1serv.saveFrameToExcel(options, looTrainFrame, excelwriter, 
        mynames.lootrainsheetname, actionName='lootrain')
    # if bsTrainFrame is not None:
    nn1serv.saveFrameToExcel(options, bsTrainFrame, excelwriter, 
        mynames.bstrainsheetname, actionName='bstrain')
        
    if 'test' in options.actionlist and options.test:
        
        message0 = "launch test?" if options.notrain else  "continue and launch test?"
        message = ""
        valdef = int(nn1serv.dialogdefaults['action']['run_test'])
        if not yesNoQuestion(message0, valdef, message, forcedefault=options.yes, doprint=doprint):
            doprint("aborted")
            finished(tempres)
            return None, None  # basicTrainTestAction
        if options.verbose > 0:
            doprint("Running Test")
        cfgfile = cfgfile if cfgfile else options.cfgfile 
        datacount = 0 if trainFrame is None else trainFrame.shape[0]
        testFrame, testresultColumn = \
        nn1serv.testAction(options, testDataFrame, datacount, IDlist, 
            cfgfile, gmxfile, dico, doprint=doprint)
            
        if options.verbose > 0:
            doprint('Test done')
    
        if options.verbose >= 5:  
            formatters = {'RoundoffTest': roundoff_format(), 'ID': IDformat()}
            doprint(testFrame.to_string(float_format=float_format(options.verbose), formatters=formatters))
            
        nn1serv.saveFrameToExcel(options, testFrame, excelwriter, 
            mynames.testsheetname, actionName='test')  
    
    startwithlootest = True
    if 'loo'in options.actionlist:
        startwithlootest = False
        
        res = startExtraAction(options, config, module, trainDataFrame, 
                tempres, actiontype='loo', excelwriter=excelwriter, 
                doprint=doprint, mynames=mynames) 
        if res is None: 
            return None, None  # basicTrainTestAction
        module, datacount, datasize, dico, gmxfile, startwithbstest = res
        
        res = nn1serv.looTrainAction(options, module, trainDataFrame, 
                resultFrame, doprint=doprint, mynames=mynames) 
        looTrainFrame, lootrainfailed, LOOResultTrainColumn, gmxfileLOO, \
            sqlLOOFileName = res
            
#         decorname = "{0}train".format(fullActionNames[options.lastaction])
        decorname = fullActionNames[options.lastaction]
        nn1serv.saveFrameToExcel(options, looTrainFrame, excelwriter, 
            mynames.lootrainsheetname, decorname)
        if options.verbose > 0:
            doprint("Writing LOO model file: {0}".format(make_extern(gmxfileLOO, options)))
            doprint("{0} done".format(fullActionNames[options.lastaction]))
        if lootrainfailed:
            looTrainFrame = None                
    if 'lootest' in options.actionlist:
        if startwithlootest:
            res = startExtraAction(options, config, module, trainDataFrame, 
                testDataFrame, apiTrainRes, tempres, actiontype='lootest', 
                excelwriter=excelwriter,
                currentTrainFrame=trainDataFrame, 
                sheetname=mynames.lootrainsheetname, resfile=resfile, 
                doprint=doNothing) 
            if res is None: 
                return None, None  # basicTrainTestAction
            module, _, looTrainSheetName, ResultTrainColumn,\
                IDList, startsloowithtest = res
            
            LOOResultTrainColumn = sum([len(val) for val in options.datafields])
            if not 'identifiers' in options.grouplist:
                LOOResultTrainColumn += 1
             
        looTestFrame = nn1serv.looTestAction(options, testDataFrame, resfile, 
            IDlist, gmxfileextra=gmxfileLOO ,doprint=doprint) #tempres, excelwriter, lootrainsheetname, apiTrainRes, 
        nn1serv.saveFrameToExcel(options, looTestFrame, excelwriter, 
            mynames.lootestsheetname)  #, with index='identifiers' in options.grouplist)
        if options.verbose > 0:
            doprint("{0} done".format(fullActionNames[options.lastaction])) 
        
    startwithbstest = True
    if 'bootstrap' in options.actionlist:
        startwithbstest = False
#         curaction = 'bootstrap'
        res = startExtraAction(options, config, module, trainDataFrame, 
                tempres, actiontype='bootstrap', excelwriter=excelwriter, 
                doprint=doNothing)
        if res is None: 
            return None, None  # basicTrainTestAction 
        module, datacount, datasize, dico, gmxfile, startwithbstest = res
        
        bsTrainFrame, bstrainfailed, BSResultTrainColumn, gmxfileBS, \
            sqlBSFileName = \
            nn1serv.bsTrainAction(options, module, trainDataFrame, resultFrame, 
                                 doprint=doprint, mynames=mynames)
        decorname = "{0}train".format(fullActionNames[options.lastaction])
        nn1serv.saveFrameToExcel(options, bsTrainFrame, excelwriter, 
            mynames.bstrainsheetname, decorname)
        if options.verbose > 0:
            doprint("{0} train done".format(fullActionNames[options.lastaction])) 
        if  bstrainfailed:
            bsTrainFrame = None                     
    
    if 'bootstraptest' in options.actionlist:
        if startwithbstest:
            module, bsTrainFrame, bstrainsheetname, trainresultColumn, IDList, \
            startsbswithtest = \
            res = startExtraAction(options, config, module, trainDataFrame, 
                testDataFrame, apiTrainRes, tempres, actiontype='bootstraptest', 
                excelwriter=excelwriter, currentTrainFrame=bsTrainFrame, 
                sheetname=mynames.bstrainsheetname, resfile=resfile, 
                doprint=doNothing)
            if res is None: 
                return None, None  # basicTrainTestAction
            module, bsTrainFrame, bstrainsheetname, trainresultColumn, IDList, \
            startsbswithtest = res
    
        bsTestFrame = nn1serv.bsTestAction(options, testDataFrame, resfile, 
            IDlist, gmxfileextra=gmxfileBS ,doprint=doprint) 
        nn1serv.saveFrameToExcel(options, bsTestFrame, excelwriter, 
            mynames.bstestsheetname) 
        if options.verbose > 0:
            doprint("{0} test done".format(fullActionNames[options.lastaction])) 

    if 'usage' in options.actionlist:
        nn1serv.usageAction(options, options.origin, options.data)

    
    if options.dotimer and (len(options.multihidden) > 1): 
        delta = float(time.time() - starttimeloc)
        doprint("Running time for {0} hidden neurons: {1}".format(
            options.hidden, second2str(delta)))

    
    libfilename = options.modulename
    if not libfilename and cfgfile:
        config = defaultedConfigParser(cfgfile)
        libfilename = config.get(root, "module")
    if not os.path.exists(libfilename):
        libfilename = os.path.join(options.savedir, libfilename)
    excelfilename = nn1serv.getExcelFileName(options)
    options.metaphorResult.append(excelfilename)
    dico = OrderedDict()
    for key, value in nn1serv.dicoversion.items():
        dico["{} version".format(key)] = (value, 'text')
    if sqlFileName:
        dico["SQL file"] = (make_extern(sqlFileName, options), 'text')
    if sqlLOOFileName:
        dico["SQL LOO file"] = (make_extern(sqlLOOFileName, options), 'text')
    if sqlBSFileName:
        dico["SQL BS file"] = (make_extern(sqlBSFileName, options), 'text')
    if gmxfile:
        dico["Model file"] = (make_extern(gmxfile, options), 'text')
    if gmxfileLOO:
        dico["Model LOO file"] = (make_extern(gmxfileLOO, options), 'text')
    if gmxfileBS:
        dico["Model BS file"] = (make_extern(gmxfileBS, options), 'text')
        
    gmCustomProperties = getCustomDict(options, dico)
    excelresult = ""
    if options.copytype and libfilename: 
        sheetList = [mynames.resultsheetname, mynames.trainsheetname, 
            mynames.testsheetname, mynames.lootrainsheetname, 
            mynames.lootestsheetname, mynames.bstrainsheetname, 
            mynames.bstestsheetname] 
        frameList = [resultFrame, trainFrame, testFrame, looTrainFrame, 
            looTestFrame, bsTrainFrame, bsTestFrame]
        modeldic, _ = ReadModule(libfilename, pptyname=options.pptyname, 
            dodelete=True)
        modeldic['bestCount'] = sorted(options.multimaxrec)
        excelresult = finishExcel(excelwriter,
            options, 
            gmCustomProperties, 
            sheetList, 
            frameList, 
            modeldic, 
            libfilename, 
            IDlist, 
            trainresultColumn,  
            testresultColumn, 
            LOOResultTrainColumn,
            loowidth, 
            traintype, 
            tempres)
        if excelresult and cfgfile and nn1serv._hasTrain(options):
            # on ne sauvegarde que si on vient d'apprendre.
            config = defaultedConfigParser(cfgfile)
            config.set('training', 'excelresult', excelresult)
            config.save()
    elif options.verbose > 0:
        doprint("no copy")

    return modeldic, gmCustomProperties, excelresult # basicTrainTestAction
    
def manageApiOptionsResult(options, toprint, stopcode, doprint, encoding):
    """Processing special asking (stop, display) in options.
    Return True to stop the program.
    """
    if stopcode & nnu.OPTION_STOP:
        doprint(toprint, file=sys.stderr)
        return 1      
    
    if stopcode & nnu.OPTION_DISPLAY_FILE_AND_STOP:
        # affichage la licence ou de l'aide prise dans le fichier "toprint"
        #if (ou tfile is not None):
        if int(options.verbose) >= 5:
            doprint(toprint, file=sys.stderr)
#            doprint()
            doprint(encoding)
            doprint()
        with open(toprint, 'rb') as ff:  #, encoding='UTF-8'
            line = ff.readline()
            while line:
                #val = line.decode("ISO-8859-15") 
                val = line.decode("UTF-8") 
                doprint(val.rstrip(), file=sys.stderr) 
                line = ff.readline()
            # val = ff.read().decode("ISO-8859-15") 
        doprint()
        return 2
    
    elif stopcode & nnu.OPTION_WARN_AND_STOP: 
        # affichage du message toprint   
        doprint(toprint, file=sys.stderr)
        return 3
    
    if options.verbose >= 5:
        import platform
        doprint("Python", platform.python_version())               
    
    if toprint:
        # affichage du message
        doprint(toprint)
        if stopcode & nnu.OPTION_WARN_AND_ASK: 
            # test si possibilite d'arret
            message = "corrected options accepted"
            valdef = int(nn1serv.dialogdefaults['action']['continue_anyway'])
            if not yesNoQuestion("continue anyway?", 'y', message, 
                forcedefault=options.yes, doprint=doprint):
                doprint("aborted")
                finished("")
                return 4
    return 0
    
def startRun(argv=[], externaloptions=None, tolink=True, nofile=False, 
        internalEnv=0, debug=0):
    if tolink:
        targetfile = os.path.expanduser("~/linkedServices.txt")
        with open(targetfile, 'w')  as ff:
            nn1serv.linkToServices(file=ff, nn1serv=nn1serv)
    # traitement de l'environnement interne        
    if internalEnv:
        if internalEnv ==1:
            target = 'arguments.txt'
        else:
            target = 'arguments{}.txt'.format(internalEnv)
        if debug:
            print('internalEnv {}'.format(target))
        argsfile = ""
        argsfile0 = os.path.expanduser(os.path.join('~', 'argsdir', target))     
        if os.path.exists(argsfile0):
            argsfile = argsfile0
        else:
            argsfile1 = os.path.expanduser(os.path.join('~', 'docker', 
                                                        'argsdir', target))
            if os.path.exists(argsfile1):
                argsfile = argsfile1
            else:
                print("cannot find argument file", argsfile0, "or", argsfile1)
        if not os.path.exists(argsfile):
            print("demo{} argument is not recognized a valid command".format(
                internalEnv))
            return  # runOnce
        setEnvironFromFile(argsfile)
        if debug:
            print("argsfile", argsfile)
            print("argv", argv)
            
    options, toprint, stopcode = nn1serv.api_options(argv, 
        proposedOptions=externaloptions, nofile=nofile, debug=debugrun)
    return options, toprint, stopcode
        
def runOnce(argv=[], externaloptions=None, doprint=False, tolink=True, 
            nofile=False, internalEnv=0):
    global nn1serv

    starttime = time.time()
    excelactionlist = ['train', 'test', 'loo', 'lootest', 'bootstrap', 
                       'bootstraptest']   
    mynames = Namespace(**namesdict)
    
    tempres = ""
    toprint = ""
    stopcode = False
    
    nn1serv = importlib.import_module("metaphor.nn1.api._api_service")
    demodebug = nn1serv.demodebug & nn1serv.DEMO_DEBUG_DEMO
    options, toprint, stopcode = startRun(argv=argv, 
        externaloptions=externaloptions, nofile=nofile, tolink=tolink, 
        internalEnv=internalEnv, debug=demodebug)
    
    if debugrun or demodebug:
        for ind, val in enumerate(argv):
            print("argv[{0}]: {1}".format(ind, val))
    
#     if internalEnv:
#         if internalEnv ==1:
#             target = 'arguments.txt'
#         else:
#             target = 'arguments{}.txt'.format(internalEnv)
#         if demodebug:
#             print('internalEnv {}'.format(target))
#         argsfile = ""
#         argsfile0 = os.path.expanduser(os.path.join('~', 'argsdir', target))     
#         if os.path.exists(argsfile0):
#             argsfile = argsfile0
#         else:
#             argsfile1 = os.path.expanduser(os.path.join('~', 'docker', 
#                                                         'argsdir', target))
#             if os.path.exists(argsfile1):
#                 argsfile = argsfile1
#             else:
#                 print("cannot find argument file", argsfile0, "or", argsfile1)
#         if not os.path.exists(argsfile):
#             print("demo{} argument is not recognized a valid command".format(
#                 internalEnv))
#             return  # runOnce
#         setEnvironFromFile(argsfile)
#         if demodebug:
#             print("argsfile", argsfile)
#             print("argv", argv)

    try:
        options.debug |= nn1serv.demodebug
    except: pass
    try:
        outfile = options.outfile
    except:
        outfile = sys.stderr
    doprint = toPrint(outfile) 
    if outfile:
        encoding = outfile.encoding  
        
    if manageApiOptionsResult(options, toprint, stopcode, doprint, encoding):
        # il y a eu une erreur ou une demande d'affichage, On sort du programme.
        return  # runOnce               
    
    if 'usage' in options.action:
        # demande d'usage
        return nn1serv.usageAction(options) # runOnce 
    
    if not options.source and 'train' in options.actionlist:
        raise nn1serv.monalError("This program needs a data file.")
    elif not options.source and not options.test and \
        'test' in  options.actionlist:
        raise nn1serv.monalError("This program needs a test file.") 
    
    createArgumentFile = 'analyse' in options.actionlist or \
        ('preproc' in options.actionlist and len(options.actionlist)==1)
        # Lecture et analyse d'un nouveau fichier de données, creation du 
        # fichier d'arguments

    # analysisAction est toujours appelée
    
    options = nn1serv.analysisAction(options, doprint=doprint, 
        forceAnalyse=createArgumentFile)
    if options is None:
        doprint("Input error. Program stopped !!!")
        return -1  # runOnce  
    
    if createArgumentFile:
        if options.notest:
            forceAction = "train"
        else:
            forceAction = "train,test"
        res = nn1serv.saveOptionsToFile(options, forceAction=forceAction, 
            forceYes=True, forceModel=True, doprint=doprint)
        return res  # runOnce
#######################################################################
    trainDataFrame = None
    filetype = options.filetype
    grouplist = options.grouplist
    if nn1serv._hasTrain(options):
        datafile = options.source
        datarange = options.datarange
        fieldlist = options.datafields
        drop = ('identifiers' in grouplist) and \
            not grouplist.index('identifiers') and (len(fieldlist[0]) == 1)
        trainDataFrame, _ = get_modelDataframe(datafile, datarange, fieldlist, 
            grouplist, filetype=filetype, drop=drop, withindexdict=False) 

    testDataFrame = None
    if nn1serv._hasTest(options):
        if not options.test:
            options.test = options.source
        if not options.testrange:
            options.testrange = 'TEST'
        if not options.testfields:
            options.testfields = options.datafields
        testfile = options.test
        testrange = options.testrange
        fieldlist = options.testfields
        testfiletype = options.testfiletype
        drop = ('identifiers' in grouplist) and \
            not grouplist.index('identifiers') and (len(fieldlist[0]) == 1)
        try:
            testDataFrame, _ = get_modelDataframe(testfile, testrange, fieldlist, 
                grouplist, filetype=testfiletype, drop=drop, 
                withindexdict=False) 
        except AttributeError as err:
            if drop:                
                fieldlist = fieldlist[1:]
                newfieldlist = [[val - 1 for val in lst] for lst in fieldlist]
                grouplist = grouplist[1:]
                drop = ('identifiers' in grouplist) and \
                    not grouplist.index('identifiers') and \
                    (len(fieldlist[0]) == 1)
                testDataFrame, _ = get_modelDataframe(testfile, testrange, 
                    newfieldlist, grouplist, filetype=testfiletype, drop=drop, 
                    withindexdict=False)
            else:
                testDataFrame = None
        except Exception as err:
            raise
#         except ValueError as err:
#             raise dataReadingError("Empty data in line {}".format(ind))
#             raise nn1serv.monalError("Error while reading test data")
            
        
    if not 'usage' in options.action:
        if not trainDataFrame:
            options.propertyName = testDataFrame.fieldsOf('outputs')[0]
            options.propertyunit = testDataFrame.units[options.propertyName]
            options.inputcount = len(testDataFrame.fieldsOf('inputs'))
        # 
            origin = nn1serv.getOrigin(options)
            if origin and os.path.isfile(origin):
                filename = origin 
                
        else:
            options.propertyName = trainDataFrame.fieldsOf('outputs')[0]
            options.propertyunit = trainDataFrame.units[options.propertyName]
            options.inputcount = len(trainDataFrame.fieldsOf('inputs'))
    
    options.tableName = 'trainingResult'
    options.inoutcount = 0
    memtarget = options.targetfolder[:]
    options.metaphorResult = []
    for seed in options.multiseed:
        # boucle sur les graines aleatoires (seed)
        excelresult = ""
        excelwriter = None
        cfgfile = ""
        resfile = ""
        trainfailed = False
        dico = None
        resultFrameList = []
        if seed < 0:
            seed = randint(0x10, 0xFFFFFFFF)    
        options.seed = seed
        if len(options.multihidden) > 1:
            # si on est en multihidden, on cree un repertoire ici pour la seed 
            # en cours
            st = strintlist(options.multihidden, delimiter="-", 
                groupdelimiter="--", compact=True)
            dirname = "{0}-SEED{1}-{2}N".format(options.root, seed, st)
            options.targetfolder = os.path.join(memtarget, dirname)
            options.targetfolder = IndexFilename(options.targetfolder)
            if not os.path.exists(options.targetfolder):
                os.makedirs(options.targetfolder)
            if options.arg_file:
                targetfile = os.path.join(options.targetfolder, "arguments.txt")
                with open(options.arg_file, 'r') as sf:
                    with open(targetfile, 'w') as tf:
                        for line in sf:
                            if len(line) and not line.startswith("#"):
                                lst = line.split("=")
                                if (len(lst) > 1) and lst[0].strip() == "SEED":
                                    lst[1] = str(seed)
                                    if line.endswith("\n"):
                                        lst[1] += "\n"  
                                line = "=".join(lst)
                            tf.write(line)

        options.configstrMem = options.configstr
        for options.hidden in options.multihidden:  # 
            # boucle sur les neurones caches (hidden)
            newtestframe = None if testDataFrame is None else \
                testDataFrame.copy()
            try:
                modeldic, gmCustomProperties, excelResult = basicTrainTestAction(options, 
                    trainDataFrame, newtestframe, doprint, resultFrameList, 
                    excelactionlist, mynames)
            except Exception as err:
                raise
        # fin boucle sur hidden

        if len(options.multihidden) > 1:
            excelResult = nn1serv.excelRecapitulation(options, resultFrameList, 
                modeldic, gmCustomProperties, doprint)
                
    # fin boucle sur seed

    if options.dotimer: 
        delta = float(time.time() - starttime)
        doprint("Running time {0}".format(second2str(delta)))
    finished(tempres)
    return excelResult  # runOnce 
    # end runOnce               

def run(argv=[], doprint=True, internalEnv=0):
    if debugrun:
        print("inside run:")
#     now = time.strftime("%c")
    try:
        res = runOnce(argv, doprint=doprint, internalEnv=internalEnv)
    except QuestionInterrupt:
        print("User interruption")
        res = None
    except Exception as err:
        res = None
        print(err)
        
    if isinstance(res, tuple) and len(res) >= 2:
        options = res[0]
        argname = res[1]
        if argname and os.path.exists(argname):
            mess = "do you want to run the new project ?"
            valdef = int(nn1serv.dialogdefaults['action']['run_new_project'])
            if yesNoQuestion(mess, default=valdef, outfile=sys.stdout, 
                forcedefault=options.yesyes):  #, verbose=verbose)
                argumentfile = "@{0}".format(argname)
                options.modelcreation = False
                try:
                    res = runOnce(argv=["", argumentfile], 
                              externaloptions=options, tolink=False)
                except Exception as err:
                    res = None
                    print(err)
    return res
    
def supervisor(args=None, doclean=True):  #, doclean=False
    if doclean:
        locenvinit = dict(os.environ)
    try:
        res = 1
        caller = os.getenv("CALLER", 0)
        argv = args if args is not None else sys.argv[1:] 
        if debugrun:
            print("supervisor entry point") 
            for val in argv:
                print("\t", val)
        firstarg = None if not len(argv) else str(argv[0])  #.lower()       
        if firstarg is None:
            # pas d'argument
            res = run(argv)
        else:
            if debugrun:
                print("firstarg", firstarg)
            if firstarg.startswith('demo'):  # and caller == 2
                # run demo
                ends = firstarg[len('demo'):] 
                if not len(ends):
                    index = 1
                    print('Run demo')
                else:
                    index = int(ends)
                    print('Run demo {}'.format(index))
                res = run(argv[1:], doprint=True, internalEnv=index)
            elif firstarg.startswith('@'):
                # fichier d'arguments en premier
                setEnvironFromFile(firstarg[1:])
                res = supervisor(argv[1:], False)
            elif firstarg.startswith('-'):
                # arguments en ligne de commande
                res = run(argv) 
            elif firstarg in ['version', 'license', 'help', 'description', 'env', 
                              'vba']:
                # information specifiques
                res = printInfo(firstarg, caller)
            elif firstarg in ['use', 'usage']:
                # mode utilisation
                res = usage(argv[1:])
            elif firstarg in ['get', 'getproperty']:
                # mode utilisation interne
                if caller == 2:
                    res = usageInterne(argv[1:])
                else:
                    res = usage(argv[1:])
            elif firstarg == 'run':
                # run standard
                res = run(argv[1:])
            elif firstarg == 'server':
                try:
                    from metaphor.metaphor_api_server import api_server
                    res = api_server.run(argv[1:])
                except Exception as err:
                    print(err)
                    print("cannot open metaphor server") 
                    res = -1
            else:
                print("'{}' argument is not recognized as a valid command".format(
                    firstarg)) 
                res = -1
    finally:
        if doclean:
            recupenv(locenvinit)
    return res  
              
if __name__ == '__main__':

    res = supervisor()
#    print("Metaphor done")


'''
Created on 10 févr. 2019

@author: jeanluc
'''
import os, sys

from argparse import ArgumentParser, FileType

from ....nntoolbox.filetoolbox import getDictFromGmx, findFile
from ....nntoolbox.utils import make_intern, DEMO_DEBUG_USAGE, isFloatEx, \
    isFloat, set2Float, getListFromStr

from .. import _api_service as nn1serv

# DEBUG_USE = 1  # Attention ici

def getCommandLineUsageOptions(args):
    caller = int(os.environ.get('CALLER', "0"))
    
    parser = ArgumentParser(description='Metaphor.', fromfile_prefix_chars='@') 

    parser.add_argument('-d', '--data', type=str, default="", dest="usagedata",
        nargs='+', help="Data to be evaluated. semicolon separated list 'id;inputs'")

    parser.add_argument('-sd', '--datarange', default="", 
        help="Range for training data (Excel-like files)")

    parser.add_argument('-sf', '--datafields', default=[], 
        help="Range for training data (Excel-like files)")

    parser.add_argument('-gl', '--grouplist', default=[], 
        help="lit of data field appli types")

    parser.add_argument('-db', '--debug', default=0, help="Debug state. Default 0")
    
    parser.add_argument('-L', '--leveragethreshold', type=float, default=0,
        help="leverage selection threshold. Default 0.0")

    parser.add_argument('-de', '--decor', nargs='?', default="t,i,e,s,se,lt", 
        help="decorator rules for result files. Default ''")

    parser.add_argument('-v', '--verbose', default="3,0", type=str, 
        help="Verbose level")
    
    usageOptions = parser.parse_args(args)
    usageOptions.curaction = ""
    usageOptions.lastaction = ""
    usageOptions.lastactions = []
    usageOptions.yesyes = True
    #usageOptions.grouplist = []
    usageOptions.groupcount = []
    usageOptions.filetype = ""
    usageOptions.actionlist = ['analyse', 'usage']
    usageOptions.yes = True
    usageOptions.configstr = ""
    usageOptions.copyexcel = 1
    nn1serv._initiate_folders(usageOptions, caller)
    usageOptions = nn1serv.initiate_externpath(usageOptions, withdialog=False)
    usageOptions = nn1serv.initiate_verbose(usageOptions)
    if isinstance(usageOptions.decor, str):
        usageOptions.decor = usageOptions.decor.split(",")
    usageOptions.decor = [val.strip() for val in usageOptions.decor]
    usageOptions.decor = nn1serv.decor2long(usageOptions.decor)
    if not usageOptions.usagedata:
        datafile = ""
    else:
        datafile = findFile(usageOptions.usagedata[0], ['~/docker', '~/docker/data', '~/docker/argsdir', '~/docker/result'])
    if isinstance(usageOptions.grouplist, str):
        usageOptions.grouplist = usageOptions.grouplist.split(',')
    if isinstance(usageOptions.usagedata, list) and bool(len(usageOptions.usagedata)):
        if os.path.isfile(datafile):
            usageOptions.source = datafile
            usageOptions.usagedata = ""
    else:
        if os.path.isfile(datafile):
            usageOptions.source = datafile
            usageOptions.usagedata = ""
    usageOptions, toprint, errorcode = nn1serv._initiate_debug(usageOptions)
    if hasattr(usageOptions, 'datafields') and isinstance(usageOptions.datafields, str):
        usageOptions.datafields = getListFromStr(usageOptions.datafields, double=True)
    
    return usageOptions

def usageInterne(cmd_line):
    internalFolder = '/home/docker/workdir'
    if not os.path.exists(internalFolder):
        internalFolder = '/Users/jeanluc/docker_work/Python3-Metaphor/metaphor-demo'
    if not len(cmd_line):
        print('need a command line')
        return
    
    iphenindex = len(cmd_line)
    for ind, val in enumerate(cmd_line):
        if val.startswith('-') and not isFloatEx(val):
            iphenindex = ind
            break

    hasdataline = '-d' in cmd_line
    index = -1
    if hasdataline:
        index = cmd_line.index('-d')
        data = cmd_line[index + 1]
        data = data.split(';')
    else:
        data = set2Float(cmd_line[:iphenindex])
        newdata = []
        ind = 0
        while ind < len(data) and isinstance(data[ind], str):
            newdata.append(data[ind])
            ind += 1
        if ind < len(data):
            lstst = [str(val) for val in data[ind:]]
            newdata.append(",".join(lstst))
#         data = newdata
        cmd_line = ["-d", ";".join(newdata)] + cmd_line[iphenindex:]
    if len(data) > 3:
        is_nn = True
    elif len(data) == 1:
        is_nn = isFloat(data[0])
    else:
        is_nn = isFloat(data[1])
    if is_nn:
        test = 'NN_'
    else:
        test = 'GM_'
    modelfile = ""
    for val in os.listdir(internalFolder):
        filetest = os.path.join(internalFolder, val)
        if os.path.isfile(filetest) and val.startswith(test) and val.endswith('.nnx'):
            modelfile = filetest
            break
    return usage([modelfile] + cmd_line)

def usage(cmd_line):
#     if DEBUG_USE:
#         print("usage command line\n\t{}".format(cmd_line))
    targetfile = os.path.expanduser("~/linkedServices.txt")
    with open(targetfile, 'w')  as ff:
        nn1serv.linkToServices(file=ff)
#     print("starting usage")
    iphenindex = len(cmd_line)
    for ind, val in enumerate(cmd_line):
        if val.startswith('-') and not isFloatEx(val):
            iphenindex = ind
            break
    cmd_line_start = cmd_line[:iphenindex]   
    cmd_line_end = cmd_line[iphenindex:]
    
    if not iphenindex:
        print("usage needs arguments")
        return
    
    modelfile = make_intern(cmd_line_start[0])
    mode = 'nn'
    hidden = 0
    if iphenindex:
        if modelfile.startswith('@'):
            argfile = modelfile[1:]
            print("fichier d'arguments : {}".format(argfile))
            print("Qu'en faire ?!?")
            return
        old = modelfile
        modelfile = findFile(modelfile, ['~/docker', '~/docker/data', '~/docker/argsdir', '~/docker/result'])
#         print("model file : {}".format(modelfile))
        if not modelfile:
            print("cannot open file : '{}'".format(old))
            return
        dico = getDictFromGmx(modelfile)
#         if not dico.get('xml', "") and not dico.get('modeldata', None):
        dicogene = dico.get('general', None)
        if dicogene and dicogene.get('grademax', 0):
            mode = 'gm'
        dicomodel = dico.get('model', None)
        if dicomodel:
            hidden = dicomodel.get('hidden', 0)       
    
    if len(cmd_line) < 2:
        print(dico)
    else:        
        usageOptions = getCommandLineUsageOptions(cmd_line_end)
        if not hasattr(usageOptions, 'propertyname') or not usageOptions.propertyname:
            usageOptions.propertyname = dico.get('propertyname', "")
        if not hasattr(usageOptions, 'test') and hasattr(usageOptions, 'source'):
            if os.path.exists(usageOptions.source):
                usageOptions.test = usageOptions.source
            if hasattr(usageOptions, 'datarange'):
                usageOptions.testrange = usageOptions.datarange 
            if hasattr(usageOptions, 'datafields'):
                usageOptions.testfields = usageOptions.datafields       
        if usageOptions.debug & DEMO_DEBUG_USAGE:
            print("usage command line\n\t{}".format(cmd_line))
        usageOptions.hidden = hidden
        usageOptions.mode = mode 
        data = usageOptions.usagedata
        nn1serv.usageAction(usageOptions, modelfile, data)
        data = ";".join(cmd_line_start[1:])

if __name__ == '__main__':
    usage(sys.argv[:1])


'''
**Application programming interface for metaphor modeling engine.**

This interface is able to manage standard Neural Networks model as well as Chemical Graphmachine models.

Preliminary release with minimum usage functionalities.

@author: Jean-Luc PLOIX
'''



#-*- coding: ISO-8859-15 -*-
# Metaphor

import os, sys, site
import ez_setup
ez_setup.use_setuptools()

from setuptools import setup, find_packages
from setuptools.command.install import install as _install
from setuptools.extension import Extension

from distutils.dir_util import mkpath
from distutils.util import convert_path
from datetime import datetime
try:
    from Cython.Build import cythonize
except: pass
# sys.stderr.write("cmdline : " + " ".join(sys.argv) + "\n")
# sys.stdout.write("cmdline : " + " ".join(sys.argv) + "\n")

def get_option(args, option_name, short_option_name=""):
    test = "--{0}".format(option_name)
    res = test in args
    if res:
        args.pop(args.index(test))
    else:
        if not short_option_name:
            short_option_name = option_name
        test = "-{0}".format(short_option_name)
        res = test in args
        if res:
            args.pop(args.index(test))
        else:
            for ind, val in enumerate(args):
                if val.startswith(test):
                    _, res = val.split('=')
                    args.pop(ind)
                    break         
    return res

debug = get_option(sys.argv, 'debug')
is_test = get_option(sys.argv, 'test')
no_cython = get_option(sys.argv, 'no-cython')
keep_C = get_option(sys.argv, 'keep-c')
no_rename_py = get_option(sys.argv, 'no-rename-py')
debug = 1

localpath = os.path.dirname(__file__)
if localpath and localpath != os.getcwd():
    os.chdir(localpath)

modulename = "metaphor"
modulepath = os.path.join(localpath, modulename)

sourcecython = ""
projectname = "metaphor"

cythonfile = os.path.join(modulename, 'cythonlist.txt')

languageLevel = int(sys.version_info[0])
    
main_ns = {}
ver_path = os.path.join(modulepath, "version.py")
ver_path = convert_path(ver_path)
try:
    with open(ver_path, encoding="ISO-8859-15") as ver_file:
        exec(ver_file.read(), main_ns)
except:
    with open(ver_path) as ver_file:
        exec(ver_file.read(), main_ns)
module_version = main_ns['__version__']

homedir = os.path.expanduser(chr(126))
saveinstallinfo = 1
debugfilemain = os.path.join(homedir, "install{0}-{1}.txt".format(modulename, module_version))
debugfilevirtual = os.path.join(homedir, "installvirtual{0}-{1}.txt".format(modulename, module_version))

def beforeRun(cythonlist):
    print("local path : {0}".format(localpath))
    cythonizelist = []
    renamelist = []
    for basefile in cythonlist:
        fullfile = os.path.join(localpath, basefile)
        pydfile = "{0}.pyd".format(fullfile)   
        pyxfile = "{0}.pyx".format(fullfile)
        pyfile = "{0}.py".format(fullfile)
        crfile = "{0}._py".format(fullfile)
        shortpyfile = "{0}.py".format(basefile)
#        shortpyxfile = "{0}.pyx".format(basefile)
        shortcrfile = "{0}._py".format(basefile)

        if os.path.exists(pyxfile):
            if os.path.exists(pyfile):
                os.rename(pyfile, crfile)
                sys.stdout.write("renaming {0} to {1}\n".format(shortpyfile, shortcrfile))
            cythonizelist.append(pyxfile)
        elif os.path.exists(pyfile):
            renamelist.append((pyfile, crfile, shortpyfile, shortcrfile))
            cythonizelist.append(pyfile)
        elif os.path.exists(crfile) and not os.path.exists(pydfile):
            os.rename(crfile, pyfile)
            sys.stdout.write("renaming {0} to {1}\n".format(shortcrfile, shortpyfile))
            renamelist.append((pyfile, crfile, shortpyfile, shortcrfile))
            cythonizelist.append(pyxfile)
    try:
        cythonize(cythonizelist, compiler_directives={'language_level' : languageLevel})
    except: pass
    ll = len(renamelist)
    for ind, (pyfile, crfile, shortpyfile, shortcrfile) in enumerate(renamelist):
        os.rename(pyfile, crfile)
        sys.stdout.write("[{2}/{3}] renaming {0} to {1}\n".format(
            shortpyfile, shortcrfile, ind+1, ll))
    return

def afterRun(projectname, cythonlist, cythonpathlist):# 
    if not keep_C:
        for basefile in  cythonlist:
            fullfile = os.path.join(localpath, basefile)
            Cfile = "{0}.c".format(fullfile)
            shortCfile = "{0}.c".format(basefile)
#            pyxfile = "{0}.pyx".format(fullfile)
            pyfile = "{0}.py".format(fullfile)
            crfile = "{0}._py".format(fullfile)
            shortpyfile = "{0}.py".format(basefile)
            shortcrfile = "{0}._py".format(basefile)
            if not no_rename_py and os.path.exists(crfile) and not os.path.exists(pyfile):
                os.rename(crfile, pyfile)
                sys.stdout.write("renaming {0} to {1}\n".format(shortcrfile, shortpyfile))
            if not no_rename_py and os.path.exists(crfile) and not os.path.exists(pyfile):
                os.rename(crfile, pyfile)
                sys.stdout.write("renaming {0} to {1}\n".format(shortcrfile, shortpyfile))
            if not 'bdist_wheel' in sys.argv and os.path.exists(Cfile):
                os.remove(Cfile)
                sys.stdout.write("deleting {0}\n".format(shortCfile))
    with open('tempfile.txt', 'w') as ff:
        ff.write(main_ns['distname'](projectname))

def getcythonlist(no_cython):
    if no_cython:
        return [], []
    cythonlist = []
    cythonpathlist = []
    cythonlistfilename = os.path.join(modulepath, cythonfile)
    if no_cython:
        cythonlist = []
    if not os.path.exists(cythonlistfilename):
        cythonlistfilename = cythonfile
    if os.path.exists(cythonlistfilename):
        with open(cythonlistfilename) as ff:
            cythonlist = [val.strip() for val in ff]
        cythonlist = [val for val in cythonlist if val and not val.startswith("#")]
    for cythonitem in cythonlist:
        lst = []
        folder, base = os.path.split(cythonitem)
        while folder:
            lst.insert(0, base)
            folder, base = os.path.split(folder)
        lst.insert(0, base)
        cythonpathlist.append(".".join(lst))
    return cythonlist, cythonpathlist

class install(_install):

    def initialize_options(self):
        self.debug = debug
        self.test = is_test
        _install.initialize_options(self)

    def finalize_options(self):
        _install.finalize_options(self)

    def isVirtual(self):
        """detect if python run in a virtual environment.
        """
        from sysconfig import get_config_var
        prefix = get_config_var('prefix')
        platbase = get_config_var('platbase')
        res = (prefix != platbase) or not hasattr(site, 'getuserbase') 
        if not res:
            try:
                prefix.index('envs')  
                res = True
            except: pass
        return res

    def get_install_path(self):
        test = self.get_outputs()[0]
        loc = ""
        while loc != modulename and test:
            test, loc = os.path.split(test)
        return test

    def doAfterInstallTest(self):
        temp = __import__(modulename)
        print(temp.__file__)
        try:
            temp.test.runtest()
        except: 
            print("No test in {0}".format(temp))
    
    def afterRunInstall(self, message, outfilename):  #, virtual, userTarget):
        def fullCythonList(localpath=""):
            if not localpath:
                return cythonlist
            if not os.path.exists(localpath):
                return []
            lst = [os.path.join(localpath, val) for val in cythonlist] 
            return lst
        
        installpath = self.get_install_path()
        sys.path.insert(0, installpath)
        installedcythonlist = os.path.join(installpath, 'metaphor', 'cythonlist.txt')
        if outfilename:
            with open(outfilename, "a") as ff:
#                 ff.write("virtual = {0}\n".format(virtual))
#                 if userTarget:
#                     ff.write("--user option ON\n")
                ff.write('installpath = {0}\n'.format(installpath))
                ff.write('localpath = {0}\n'.format(localpath))
                ff.write(message)
        if debug:
            print("{0} path : {1}".format(modulename, installpath))
        cleanlist = fullCythonList(installpath)
        for fil in cleanlist:
            ffil = '{0}{1}'.format(fil, ".c")
            if os.path.exists(ffil):
                os.remove(ffil)
                sys.stderr.write('removed {0}\n'.format(ffil))
            else:
                if debug:
                    sys.stderr.write('cannot find {0}\n'.format(ffil))  
 
        
#         print("runCompile MonalLib")
#         from metaphor.monal.util.utils import runCompil eMon alLib
#         lcodepath = os.path.join(installpath, 'metaphor', 'monal', 'lcode')
#         res = runCom pileMonal Lib(doprint=True, freesource=True, force=True, 
#             debug=self.debug, logfile=outfilename, libpath=lcodepath)
#         print("end runCo mpileMona lLib :", res)
# # remove common folder after compile lib
#         commonfolder = os.path.join(installpath, 'metaphor', 'monal', 'lcode', 'code_templates', 'common')  
#         if res and os.path.exists(commonfolder):
#             shutil.rmtree(commonfolder)
        os.remove(installedcythonlist)      
        
        if self.test:
            self.doAfterInstallTest()
 
    def beforeRunInstall(self, message):
        if self.debug:
            print("debug = {0}".format(self.debug))
        virtual = self.isVirtual()
        if virtual and "--user" in sys.argv:
            sys.argv.pop(sys.argv.index("--user"))
        userTarget = "--user" in sys.argv 
        outfilename = "" if not saveinstallinfo else debugfilevirtual if virtual else debugfilemain
        if outfilename:
            with open(outfilename, "w") as ff:
                ff.write(datetime.now().strftime("%d/%m/%Y %H:%M:%S\n"))
                ff.write("virtual = {0}\n".format(virtual))
                if userTarget:
                    ff.write("--user option ON\n")
                ff.write(message)
            return outfilename  #, virtual, userTarget
        return ""

    def run(self):
        pre_install_message = "start original {0} install\n".format(modulename)
        outfilename = self.beforeRunInstall(pre_install_message)                
        res = _install.run(self) 
        post_install_message = "original {0} install done\n".format(modulename)
        self.afterRunInstall(post_install_message, outfilename)
        return res

def runsetup(projectname, no_cython, cythonlist, cythonModlist):

    licence = """Apache License Version 2.0, January 2004
    http://www.apache.org/licenses/
    """
    cythonClistExt = ["{0}.c".format(val) for val in cythonlist]
    try:
        with open('README.md', 'r') as ff:
            long_Description = ff.read()
    except:
        long_Description = ""

    extension_params = [(valname, val) for (valname, val) in zip(cythonModlist, cythonClistExt)]
    extmodules = [Extension(valname, [val]) for valname, val in extension_params]

    installdev_url = "https://pypi.python.org/pypi/{0}".format(modulename)
    excludelist = ["metaphor.nntoolbox", "metaphor.nntoolbox.toktoolbox"]
    packages = find_packages(exclude=excludelist)
    entrypoints = {'console_scripts': [
                                       "supervisor = metaphor.nn1.api.run.nn_run:supervisor",
                                       "usage = metaphor.nn1.api.use.nn_use:usage",
                                       "linkservices = metaphor.nn1.api._api_service:linkToServices",
                                       "license = metaphor.nn1.api.printfile:printlicense",
                                       "version = metaphor.nn1.api.printfile:printversion",
                                       "help = metaphor.nn1.api.printfile:printhelp",
                                       "description = metaphor.nn1.api.printfile:printdescription",
                                       "complib = metaphor.monal.util.utils:docompilelib",
                                       #"compilelib = metaphor.monal.util.utils:run"
                                       ],}
    installrequires = ('six', 
                       'numpy>=1.14', 
                       'scipy>=O.19',
                       'tqdm', 
                       'pandas>=0.22.0', 
                       'xlrd>=1.0', 
                       'xlwt>=1.2', 
                       'openpyxl>=2.5.0',
                       'metaphor_nntoolbox')

    classifiers = [ "Development Status :: 4 - Beta",
                    "Environment :: MacOS X",
                    "Intended Audience :: Science/Research",
                    "License :: OSI Approved :: Apache Software License",
                    "Operating System :: MacOS :: MacOS X",
                    "Programming Language :: Python :: 3.6",
                    "Topic :: Scientific/Engineering :: Artificial Intelligence",
                    ]

    setup(name=projectname,
          version=module_version,
          description="Neural network toolbox",
          long_description=long_Description,
          long_description_content_type="text/markdown",
          author="Jean-Luc PLOIX",
          author_email="jean-luc.ploix@espci.fr",
          url="http://www.netral.fr",
          download_url=installdev_url, 
          platforms = ['Windows', 'Linux', 'Mac OS X'],
          license = licence,
          entry_points = entrypoints,
          zip_safe=False,
          packages = packages,
          install_requires = installrequires,
          classifiers = classifiers,
          cmdclass={'install': install}, 
          ext_modules = extmodules,
          include_package_data=True,
          )

if __name__ == "__main__":

    if len(sys.argv) == 1:
        sys.argv.append("sdist")
    try:
        if "SETUPPATH" in os.environ:
            pth = os.environ["SETUPPATH"]
            if not os.path.exists(pth):
                mkpath(pth)
            os.chdir(pth)
            
    except: 
        pass

    #action d'installation
    installaction = 'install' in sys.argv or 'develop' in sys.argv
    #action de preparation des distributions
    distaction = 0
    for arg in sys.argv[1:]:
        if arg[:5] in ('sdist', 'bdist', 'build'):
            distaction = 1
            break

    cythonlist, cythonpathlist = [], []
    if not no_cython:
        print("run beforeRun")
        cythonlist, cythonpathlist = getcythonlist(no_cython)
        if distaction:
            beforeRun(cythonlist)
        if installaction and not len(cythonlist):
            sys.stdout.write("Empty cython list\n")
        print("beforeRun done")
    if is_test:
        for val, pth in zip(cythonlist, cythonpathlist):
            print("{0}\t\t{1}".format(val, pth) )            
    else:
        runsetup(projectname, no_cython, cythonlist, cythonpathlist) 

    if distaction and not no_cython:
        print("run afterRun")
        afterRun(projectname, cythonlist, cythonpathlist)
        print("afterRun done")
    if distaction:
        print("metaphor setup done")


#-*- coding: ISO-8859-15 -*-
#===============================================================================
# $Id: _node.py 4819 2018-11-02 05:32:25Z jeanluc $
#  Module  _node
#  Projet MonalPy
#
#  Implementation python de monal
#
#  Author: Jean-Luc PLOIX  -  NETRAL
#  Decembre 2009
#
# Copyright 2016 Jean-Luc PLOIX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import warnings

with warnings.catch_warnings():
    warnings.filterwarnings("ignore",category=DeprecationWarning)

from operator  import mul as opmul
from numpy import array, asarray, ndarray, zeros, zeros_like, dtype, NaN, \
    delete as npdelete
from functools import reduce

from ._monalfunc import identityactiv, getIndexFromActivClass, getActivClass
from ..Property import Property
from ..util.monalbase import Component, MonalError, DimensionError, LinkError
from ..util.utils import CCharOnlys  #, ChildElementList
from ._link import Synapse
from ..util.monaltoolbox import register
from ..util import const_st as CS
from ..monalconst import NS_STATE, NS_PI, LS_STD, LS_FIXE, LS_TELQUEL, \
    FIX_LIMIT, ID_COMPUTE, ID_DIFF, GC_MULTI_DATA, GC_MULTI_OUTPUT #, NS_INPUT
from ..monalrecords import NodeData
from ..specialmath import INF

__all__ = ["Basenode", "Node", "Nodepi", "newNode"]

candidateattr = ("atom", "mass")
# noms des attributs facultatifs qui peuvent être attribués lors de la
    #construction d'un "Node". Ces attributs sont conservés lors de la sauvvegarde XML
    
XMLRELATIVE_01 = 1  # 1 -> intervalle [0, 1]
                    # 0 -> intervalle [-1, 1]

class Basenode(Component):

    def __init__(self, owner, name="", layer=0, pos=-1, activ="", defaultvalue=0.0, style=0):
        # owner doit etre un Network
        super(Basenode, self).__init__(owner, name)
        self.style = style
        self.links = []
        self.downLinks = []
        self.layer = layer
        self.nodeid = None
        self.activation = activ
        self.value = defaultvalue
        self._tagname = 'NODE'
        self.derive = array([1.])
        self.partial = array([0.])
        self.TestNoMemSyn = True
        self._useOriginal = False
        self.computelevel = 1
        self.DLambdaZ = None
        self.DLambdaW = None
        self._mini = None
        self._maxi = None

    def __del__(self):  # Basenode
        try:
            del self._activfunc
        except: pass
        self.downLinks = []
        self.links = []
        super(Basenode, self).__del__()

    def reprlst(self):  # Basenode
        lst = super(Basenode, self).reprlst()
        lst.append("nodeid : {}".format(self.nodeid))
        lst.append("layer : {}".format(self.layer))
        lst.append("activation : {}".format(self.activation))
        lst.append("value : {}".format(self.value))
        return lst

    @Property
    def linksCount(self):  # Basenode
        return len([val for val in self.components if isinstance(val, Synapse)])

    @classmethod
    def createFromElementXML(cls, owner, element, localversion):  # Basenode
        name = element.get("neuronname", "") or element.get("name", "")
        style = int(element.get("style", "0"))
        layer = int(element.get("layer", "0"))
        funct = element.get("function", "") or element.get("func", "")
        nodeid = int(element.get("nodeid", "0"))
        res = cls(owner, name, layer, activ=funct)
        for attr in candidateattr:
            try:
                st = element.get(attr, "")
            except:
                st = ""
            if st:
                setattr(res, attr, st)
        res.style = style
        res.nodeid = nodeid
        lst = list(element)# list(element.getchildren())
        haspot = not len(lst)  #list(element.getchildren()))
        for subelmt in lst:
            if res.includeSubElement(res.links, subelmt, localversion): pass  #, weak=True
            elif subelmt.text != 'None':
                if subelmt.tag == "MINIMUM": res._mini = float(subelmt.text)
                elif subelmt.tag == "MAXIMUM": res._maxi = float(subelmt.text)
                elif subelmt.tag == "VALUE": res.value = float(subelmt.text)
                elif subelmt.tag == "POTENTIAL":
                    res.potential = float(subelmt.text)
                    haspot = True
                elif subelmt.tag == "INTERFACE":  # version 27 a 32
                    mini = subelmt.get("minimum", 0.0)
                    if (mini != "NAN") and mini is not None:
                        res._mini = float(mini)
                    maxi = subelmt.get("maximum", 1.0)
                    if (maxi != "NAN") and maxi is not None:
                        res._maxi = float(maxi)
                    try:
                        relinput = float(subelmt.get("relinput", "0"))
                        if XMLRELATIVE_01:
                            delta = (res._maxi - res._mini) * relinput
                        else:
                            delta = (res._mini * (1 - relinput) + res._maxi * (1 + relinput)) / 2
                        res.value = res._mini + delta
                    except: pass
        if not haspot:
            res.potential = res.value
        return res

    def getmaxlinkid(self, origin=True):  # Basenode
        if origin:
            lst = [syn.originalid for syn in self.links if syn.linkid < FIX_LIMIT]
        else:
            lst = [syn.linkid for syn in self.links if syn.linkid < FIX_LIMIT]
        if len(lst):
            return max(lst)
        return 0

    @Property
    def useOriginal(self):  # Basenode
        return self._useOriginal
    @useOriginal.setter
    def suseOriginal(self, value):  # Basenode
        self._useOriginal = value
        for syn in self.links:
            syn.useOriginal = value

    def IsFixed(self, fromInput):  # Basenode
    #=============================================================================
    # Result := true;
    # i := 0;
    # while Result and (i < CompteParent) do
    # begin
    #   SS := Synapses[i];
    #   Result := SS.Fixed;
    #   if Result and FromInput then
    #     Result := SS.NParent.GetIsFixed(true);
    #   inc(i);
    # end;
    #=============================================================================
        res = True
        for syn in self.links:
            res = syn.isFix()
            if res and fromInput:
                res = syn.source.IsFixed(True)
            if not res:
                break
        return res

    def outputnum(self):  # Basenode
        return self.owner.outputnum(self)

    def isPi(self):  # Basenode
        return False

    def equivalent(self, target, withException=False):  # Basenode
        eq = True
        for syn1, syn2 in zip(self.orderedLinks, target.orderedLinks):
            eq &= syn1.equivalent(syn2, withException) and (syn1.source.nodeid == syn2.source.nodeid)
            if withException and not eq:
                raise AssertionError("Different link sources %s: %s, %s: %s"%(syn1.linkid, syn1.source.nodeid, syn2.linkid, syn2.source.nodeid))
        return eq
    
    def xmlAttrList(self):  # Basenode
        res = []
        self.appendDefaultedAttribute(res, 'nodeid', self.nodeid)
        self.appendDefaultedAttribute(res, 'name', self.name)
        for attr in candidateattr:
            if hasattr(self, attr):
                self.appendDefaultedAttribute(res, attr, getattr(self, attr))
        self.appendDefaultedAttribute(res, 'style', self.style, default='0')
        self.appendDefaultedAttribute(res, 'layer', self.layer)
        self.appendDefaultedAttribute(res, 'function', self.activation, default="ident")
        return res
    
    def isLin(self):  # Basenode
        act = self._activfunc
        return (act is None) or isinstance(act, identityactiv)
    
    def xmlChildren(self, ident='\t'):  # Basenode
        res = []
        if hasattr(self, "potential") and (self.value != self.potential):
            res = ['%s<POTENTIAL>%s</POTENTIAL>'% (ident, self.potential)]
        if self.value:
            res.append('%s<VALUE>%s</VALUE>'% (ident, self.value))
        return res
            
    def getSynapse(self, source, style=LS_STD, linkid=-1, name=None):  # Basenode
        #for link in self.downLinks:
        for link in self.links:
            if ((link.source == source) and  
            ((linkid < 0) or (link._linkid == linkid)) and
            ((not name) or (link.name == name))):
                return link
        return None            
    
    def absorb(self, source, inputs=False, outputs=False):  # Basenode
        if inputs:
            while source.linksCount:
                syn = source.links[0]
                syn.owner = self
                self.links.append(syn)
        if outputs:
            while len(source.downLinks):
                syn = source.downLinks.pop(0)
                syn.source = self
    
    def isolate(self, inputs=False, outputs=False, weights=None, style=LS_STD):  # Basenode
        if inputs:
            for syn in list(self.links)[-1::-1]:
                self.deleteLink(syn, weights=weights, style=style)
        if outputs:
            while len(self.downLinks):
                syn = self.downLinks.pop()
                if not (style & LS_TELQUEL|LS_FIXE) and weights and (syn.linkid < len(weights)):
                    weights.pop(syn.linkid)     
                syn.target.links.remove(syn)
                syn.owner = None
            
#             lst = self.downLinks[-1::-1]
#             for syn in lst:
#                 
#                 syn.target.deleteLink(syn, style=style)
    
    def deleteLink(self, source, style=LS_STD, linkid=-1, name=None, weights=None):  # Basenode
        res = -1
        if isinstance(source, Synapse):
            syn = source
        else:
            syn = self.getSynapse(source, style, linkid, name)
        if syn:
            res = syn._linkid
            syn.owner.links.remove(syn)
            if not(style & (LS_TELQUEL|LS_FIXE)):
                if weights is not None:
                    if isinstance(weights, list):
                        weights.pop(syn.linkid)
                    else:
                        weights = npdelete(weights, [syn.linkid])
            try:
                syn.source.downLinks.remove(syn)
                # this remove is done at the end to keep a reference to syn 
                # until there
            except Exception:
                pass
            syn.owner = None
        return res
    
    def createLink(self, source, value=0.0, style=LS_STD, linkid=-1, name=""):  # Basenode
        """Create a new link from source toward target"""
        cursyn = Synapse(owner=self, source=source, value=value, style=style, name=name)
        self.links.append(cursyn)

        # accepter le lien et calculer son numero d'ordre, ...
        if not (LS_FIXE & cursyn.style) and (linkid <= FIX_LIMIT):
            wlst = list(self.network._weights)
            if linkid < 0:
                wlst.append(cursyn.value)
                linkid = len(wlst) - 1
            elif len(wlst) > linkid:
                for synloc in self.network.links:
                    if (synloc.linkid >= linkid) and not synloc.isFix():
                        synloc.linkid += 1
                wlst.insert(linkid, cursyn.value)
            else:
                while len(wlst) <= linkid:
                    wlst.append(NaN)
                wlst[linkid] = cursyn.value
#                 while len(self.network._weights) <= cursyn.linkid:
#                     self.network._weights.append(NaN)
#                 self.network._weights[linkid] = cursyn.value
#                 self.network._weights = asarray(wlst)
#             cursyn.linkid = linkid
            self.network._weights = asarray(wlst)
        elif linkid > FIX_LIMIT:
            self.network._lastFix = max(linkid, self.network._lastFix)
        else:
            linkid = self.network._lastFix
            self.network._lastFix += 1
        # OK, le lien est accepte        
        #self._parameternames[linkid] = syn.name
        cursyn.linkid = linkid
        return cursyn
    
    def reconnect(self, owner):  # Basenode
        # - etablissement du proprietaire
        # - branchement des liaisons pendantes
        self._owner = owner
        for syn in self.links:
            syn.reconnect(self)
    
    #def __setstate__(self, state):  # Basenode
    #    self.__dict__.update(state)
    
    def __getstate__(self):  # Basenode
        # preparation du pickling
        objdict = self.__dict__.copy()
        objdict["downLinks"] = []
        #objdict["components"] = [] 
        objdict["_owner"] = None
        return objdict
    
    def organize(self, network):  # Basenode
        assert(network)
        for syn in self.links:
            syn.organize()  #self, network
        for syn in list(self.downLinks):
            tg = syn.target
            if tg is None:
                raise LinkError("Ownerless Synapse #%d  on node #%d"% (syn.linkid, self.nodeid))
            if (tg.nodeid < self.nodeid) and (tg.style & NS_STATE):
            # syn.target est situee avant dans le reseau, et est un noeud d'etat
            # donc self est une sortie d'etat
                self.style &= NS_STATE
            
    def setMinMax(self, mini=None, maxi=None):  # Basenode
        if maxi is None:
            try: 
                maxi = mini[1]
                mini = mini[0]
            except: pass
        self._mini = mini
        self._maxi = maxi
        
    @Property
    def interval(self): # Basenode
        try:
            return (self._mini, self._maxi)
        except:
            return(-INF, INF)
    
    @Property
    def minimum(self):  # Basenode
        try: 
            val = self._mini
            if val is None:
                return -INF
            return val
        except: return -INF
            
    @Property
    def maximum(self):  # Basenode
        try: 
            val = self._maxi
            if val is None:
                return INF
            return val
        except: return INF
        
    @Property
    def mean(self): # Basenode
        min = self.minimum
        max = self.maximum
        try:
            return (min + max)/2
        except:
            return None
            
    def nameC(self):  # Basenode
        return CCharOnlys(self.name)
    
    def activationIndex(self):  # Basenode
        return getIndexFromActivClass(self.activation)
    
    def todata(self, nodedata=None):  # Basenode
        if nodedata is None:
            nodedata = NodeData()
        nodedata.name = self.name
        nodedata.layerIndex, nodedata.nodeindex = self.getIndexes()
        nodedata.activation = getIndexFromActivClass(self.activation)
        nodedata.nodeID = self.nodeid
        nodedata.style = self.style
        #nodedata.nodeType = self.nodetype
        nodedata.value = self.value    
        return nodedata
    
    def getIndexes(self):  # Basenode
        ll = self.layer
        nn = -1
        for anode in self.owner.layers(ll):
            nn += 1
            if anode == self:
                return ll, nn
        return ll, -1
    
    def defaultname(self):  # Basenode
        ll, nn = self.getIndexes()
        if nn < 0:
            return "badly inserted in network"
        if not ll:
            return CS.st_INdefault % nn
        lc = self.owner.layerCount
        if ll == lc-1:
            return CS.st_OUTdefault % nn
        ln = self.owner.layerLength(ll)
        if ln == 2:
            return "H%d"% (ll-1)
        return CS.st_HIDdefault % (ll-1, nn)
    
    def isCommmon(self, value=None):  # Basenode      
        return (len(self.links) == 1) and len(self.downLinks) \
            and self.links[0].isCommon(value=value)
    
    @Property
    def activation(self):  # Basenode
        if self._activfunc:
            return str(self._activfunc)
        return ""
    @activation.setter
    def setactivation(self, value):  # Basenode
        self._activclass = getActivClass(value)
        self._activfunc = self._activclass()

    @Property
    def source(self):  # Basenode
        ll = self.linksCount
        if not ll or (ll > 1):
            raise DimensionError("Error in source assignment")
        try: return self.links[0].source
        except: raise MonalError("Error in source assignment")
    @source.setter
    def ssource(self, value):  # Basenode
        if isinstance(value, (int, float, dtype)):#
            self.links[0].source = value
        else:
            self.links[0].source = value.flat
    
    @Property
    def orderedLinks(self):  # Basenode
        lst = (syn for syn in self.links if isinstance(syn, Synapse))
        return sorted(lst, key=lambda syn: syn.linkid)
    
    def activFormula(self, expr):  # Basenode
        return self._activfunc.formula(expr)
    
    @Property
    def network(self):  # Basenode
        return self.owner
    
    @Property
    def freeLinks(self):  # Basenode
        return (syn for syn in self.links if not syn.isFix())
        
    def adjustLinkid(self, OkMin=1000000):
        for syn in self.links:
            if syn.linkid > OkMin:
                syn.linkid -= 1
    
    def partialBase(self):
        #res = sum(syn.deltaPartial() for syn in self.links)
        res = zeros_like(self.partial)
        for syn in self.links:
            res += syn.deltaPartial()
        self.partial = res * self.derive #* syn.sousproduit
    
    def partialExtra(self):  # Basenode
        for syn in self.links:
            if syn.linkid < FIX_LIMIT:
                self.partial[syn.linkid] += self.derive * syn.source.value #* syn.sousproduit
        
    def calculDirectLocal(self, index=-1):  # Basenode
        cumul = 0
        for syn in self.links:
            if syn.linkid == index:
                cumul += syn.source.value #* syn.sousproduit
            cumul += syn.deltaPartial() 
        self.partial = cumul * self.derive   
    
    def computepotential(self):  # Basenode
        #return sum(map(lambda link: link.output(), self.links))
        return sum(link.output() for link in self.links)

    def formula(self, light=True):  # Basenode
        expr = self.potentialformula(light)
        return self.activFormula(expr)
    
    def potentialformulaList(self, light=True):  # Basenode
        lst = []
        for link in self.links:
            source = link.source
            if link.isFix():
                valstr = str(link.value)
            else:
                valstr = "{%s}" % link.name
            if source.name == "bias":
                lst.append(valstr)
            elif light:
                if valstr == "1.0":
                    lst.append(source.name)
                else:
                    lst.append("%s*(%s)" %(valstr, source.name))
            else:
                if valstr == "1.0":
                    lst.append(source.formula(light))
                else:
                    lst.append("%s*(%s)" %(valstr, source.formula(light)))
        return lst
    
    def prepareComputeHessian(self, paramCount):  # Basenode
        localnodeCount = self.linksCount
        self.DLambdaZ = zeros(self.nodeCount)
        self.DLambdaW = zeros(paramCount)
        for SS in self.links:
            SS.prepareComputeHessian(localnodeCount, paramCount)
        
    
    def propagate(self):  # Basenode
        self.potential = self.computepotential()
        self.value = self._activfunc(self.potential)
        if self.computelevel > ID_COMPUTE:
            self.derive = array(self._activfunc.derive(self.potential))
            if self.computelevel > ID_DIFF:
                self.derive2 = self._activfunc.derive2(self.potential)
        return self.value
    
    def backPropagateOld(self, gradient=None, local = None, gradDim=0, mode=0):  # Basenode
        # Calcul de la dérivée par retropropagation 
        self._lambda = local
        for syn in self.downLinks:  # balayage des enfants
            delta = syn.owner._lambda * syn.owner.derive * syn.value
            self._lambda += delta

        lambdaDerive = array([0.])
        if self.layer or (gradient is None):
            if (gradient is None):
                if mode & GC_MULTI_OUTPUT:
                    no = no = local.shape[0]
                else:
                    no = 1
                nd = 1
            elif (gradDim <= 1):
                no = 1
                nd = 1
            elif gradDim ==2:
                if mode & GC_MULTI_DATA:
                    no = 1
                    nd = gradient.shape[0]
                else:
                    if isinstance(local, ndarray) and (local is not None):
                        no = local.shape[0]
                    else:
                        no = gradient.shape[0]
                    nd = 1
            else:  # ici, gradDim = 3
                no = gradient.shape[0]
                nd = gradient.shape[1]
            if isinstance(self._lambda, (int, float, dtype)):#
                lambdaDerive = self.derive * self._lambda
            else: 
                lambdaDerive = zeros((no, nd))
                for outindex in range(no):
                    lambdaDerive[outindex, :] = self.derive * self._lambda[outindex]
            
        if (self.layer > 0) and (gradient is not None):
#            iterr = ((syn.linkidid, syn.source.value) for syn in self.freeLinks)
            iterr = ((syn.originalid, syn.source.value) for syn in self.freeLinks)
            if gradDim == 1:
                for synid, val in iterr:
                    gradient[synid] += lambdaDerive * val  
            elif gradDim == 2:
                for synid, val in iterr:
                    value = lambdaDerive * val
                    if mode & GC_MULTI_DATA:
                        try:
                            gradient[:, synid] += value
                        except:
                            gradient[:, synid] += value[0]
                    else:
                        for outindex in range(no):
                            gradient[outindex, synid] += value[outindex][0]
                          
            elif gradDim == 3: 
                for synid, val in iterr:
                    gradient[:, :, synid] += lambdaDerive * val 

        if self.computelevel > ID_DIFF:
        #  if (ComputeLevel >= 2) then
        #  begin
        #    DLambdaZ.AssignAll(0);
        #    DLambdaW.AssignAll(0);
        #    // Calcul des vecteurs DLambdas.
        #    CalculDLambda;
        #  end;
            pass  # a developper
        
        return lambdaDerive
            
    def backPropMultiOut(self, gradient=None, local = None):  # Basenode
        # Calcul de la dérivée par retropropagation 
        # la forme du gradient est (no, nd, ns) 
        self._lambda = local
        for syn in self.downLinks:  # balayage des enfants
            delta = syn.owner._lambda * syn.owner.derive * syn.value
            if self._lambda is None:
                self._lambda = delta
            else:
                self._lambda += delta

        lambdaDerive = None
        if self.layer or (gradient is None):
            if isinstance(self._lambda, ndarray):
                no = self.owner.outputCount
                nd = self.inputDim
                lambdaDerive = zeros((no, nd))
                for outindex in range(no):
                    lambdaDerive[outindex] = self.derive * self._lambda[outindex]
            else:
                lambdaDerive = self.derive * self._lambda
        
        if self.computelevel > ID_DIFF:
        #    DLambdaZ.AssignAll(0);
        #    DLambdaW.AssignAll(0);
        #    // Calcul des vecteurs DLambdas.
        #    CalculDLambda;
            pass  # a developper
        
        if not self.layer:
            return lambdaDerive.squeeze()
        
        if (gradient is not None):
            iterr = ((syn.originalid, syn.source.value) for syn in self.freeLinks)
#            iterr = ((syn.linkidid, syn.source.value) for syn in self.freeLinks)
            for synid, val in iterr: 
                if len(gradient.shape) == 2:
                    gradient[:, synid] += lambdaDerive * val
                else:
                    gradient[synid] += lambdaDerive * val 
#                gradient[:, :, synid] += lambdaDerive * val
                 
        return None

    def backPropSingleOut(self, gradient=None, local=None, original=True): # Basenode
        # Calcul de la dérivée par retropropagation pour une seule sortie 
        # la forme du gradient est (ns) 
        self._lambda = local
        for syn in self.downLinks:  # balayage des enfants
            if self._lambda is None:
                self._lambda = syn.owner._lambda * syn.owner.derive * syn.value
            else:
                self._lambda += syn.owner._lambda * syn.owner.derive * syn.value

        lambdaDerive = None
        if self.layer or (gradient is None):
            lambdaDerive = self.derive * self._lambda
        
        if (self.layer > 0) and (gradient is not None):
#            iterr = ((syn.linkid, syn.source.value) for syn in self.freeLinks)
            if original:
                iterr = ((syn.originalid, syn.source.value) for syn in self.freeLinks)
            else:
                iterr = ((syn.linkid, syn.source.value) for syn in self.freeLinks)
            for synid, val in iterr:
                if gradient.ndim == 2:
                    gradient[:, synid] += lambdaDerive * val 
                else:
                    gradient[synid] += lambdaDerive * val 

        if self.computelevel > ID_DIFF:
        #    DLambdaZ.AssignAll(0);
        #    DLambdaW.AssignAll(0);
        #    // Calcul des vecteurs DLambdas.
        #    CalculDLambda;
            pass  # a developper
        
#        if len(lambdaDerive.shape) == 2:
#            return lambdaDerive[0]
#        else:
        return lambdaDerive
                        
class Node(Basenode):
    """additive node (neuron)"""

    def potentialformula(self, light=True):  # Node
        lst = self.potentialformulaList(light)
        return " + ".join(lst)
    
    def xmlAttrList(self):  # Node
        return ['class="NeuronSigma"'] + super(Node, self).xmlAttrList()
    
class Nodepi(Basenode):
    """multiplicative node (pi neuron)"""
            
    def __init__(self, owner, name="", layer=0, pos=-1, activ=""):  # Nodepi
        # owner doit etre un Network
        super(Nodepi, self).__init__(owner, name, layer, pos, activ)
        self.TestNoMemSyn = False
        
    def potentialformula(self, light=True):  # Nodepi
        lst = self.potentialformulaList(light)
        lst = ["(%s)"% val for val in lst]
        return " * ".join(lst)
    
    def xmlAttrList(self):  # Nodepi
        return ['class="NeuronPi"'] + super(Nodepi, self).xmlAttrList()

    def computepotential(self):  # Nodepi
        return reduce(opmul, (link.output() for link in self.links))
    
    def isPi(self):  # Nodepi
        return True
    
def newNode(owner, name="", layer=0, pos=-1, activ="", style=0, defaultvalue=0.0):
    if style & NS_PI:
        return Nodepi(owner, name, layer, pos, activ, defaultvalue=defaultvalue)
    return Node(owner, name, layer, pos, activ)

register(Node)
register(Node, "Neuron")
register(Node, "NeuronSigma")
register(Nodepi)
register(Nodepi, "NeuronPi")



#-*- coding: ISO-8859-15 -*-
#===============================================================================
# $Id: monalrecords.py 4789 2018-06-08 11:40:08Z jeanluc $
#  Module _ndkWorkshop.py
#  Projet MonalDLL
#  MonaEx70 DLL interface avancé (Neuro Developer Kit)
#
#  Author: Jean-Luc PLOIX  -  NETRAL
#  July 2008
# 
#===============================================================================
"""Module defining useful 'Structure's for Neural developper kit (NDK).
"""
from ctypes import Structure, sizeof, c_short, c_ushort, c_byte, c_int, byref, \
    c_long, c_double, create_string_buffer, memset, memmove, addressof, string_at  #, c_char
from functools import partial

from .Property import Property, setter
from .util.utils import Class_doc
from . import monalconst as C

# __all__ = ["LoopData", "SynapseData", "NodeData", "LayerData", "MergeData", 
#            "ModelData", "DataDrivenModel", "DataDrivenModelTrain", "baseData"]

#-----------------------------------------------
class baseData(Structure):
    __doc__ = Class_doc("""Abstract base class for data-structures.""")
    def _setstr_at(self, field, value, lenmax=256):
        start = addressof(self) + field.offset
        val = value.encode()
        st = create_string_buffer(val)           
        memmove(start, byref(st), min(len(st), lenmax))
            
    def _getstr_at(self, field):
        start = addressof(self) + field.offset
        return string_at(start).decode()
#---------------------------------------
class LoopData (baseData):  # commun avec Delphi (TLoopDataEx)
    """
    Model loop creation data.
    
    Used for ModifyNetwork, with style: AddLoop.
    
    Fields :
        
    :layer: layer index
    :node: Node index in layer
    :state: State index
    :redondancy: Take care of redondancy while building a loop
    :notCreateSynapses: Dont create synapses
    :existingnode: Existing origin node
    """    
    #===========================================================================
    # __doc__ = Class_doc("""Donnees de bouclage reseau.
    # Utilisé pour ModifyNetwork, avec le style: AddLoop.""")
    #===========================================================================
    _fields_ = [
        ("layer", c_short),
        ("node", c_short),
        ("state", c_short),
        ("redondancy", c_ushort),
        ("notCreateSynapses", c_ushort),
        ("existingnode", c_ushort)]
    
    def __repr__(self):
        st = super(LoopData, self).__repr__()
        st2 = '(%s, %s, %s, %s, %s, %s)'% (self.layer, self.node, self.state, 
            self.redondancy, self.notCreateSynapses, self.existingnode)
        return st + st2

    def __init__(self, layer=0, node=0, order=1, nosynapse=0, existingnode=0):
        super(LoopData, self).__init__()
        memset(byref(self), 0, sizeof(self))
        self.layer = layer
        self.node = node
        self.state = order
        self.redondancy = 1
        self.notCreateSynapses = nosynapse
        self.existingnode = existingnode

#---------------------------------------
class SynapseData (baseData):  #  commun avec Delphi
            # Pour Movenode, si <> 0, l'entrée de rang
            # RangOrigine est déplacé dans la couche constante
            # avec la valeur "Valeur"
            # Pour MERGENODE si <> 0, absorbtion des synapses parents
    """
    Node link data.
    
    Used for ModifyNetwork with styles :
    
        - Movenode
        - MergeNeuron
        - NouvelleSynapse
        - DetruireSynapse
        - ModifierSynapse
        
    Fields:
    
    :size: Current record size
    :originLayer: Layer index of origin node
    :originIndex: Index of origin node in its layer
    :targetLayer: Layer index of target node
    :targetIndex: Index of target node in its layer
    :style: Link style
    :inputBool: Include input in action
    :outputBool: Include output in action
    :value: Parameter value
    :_name: Parameter name
    :netIndex: Model Index in multi model
    :index: Link index in model
    
    """
    #===========================================================================
    # __doc__ = Class_doc("""Données de lien entre noeuds.
    # Utilisé pour ModifyNetwork avec les styles :
    #    Movenode
    #    MergeNeuron
    #    NouvelleSynapse
    #    DetruireSynapse
    #    ModifierSynapse""")
    #===========================================================================
    _fields_ = [
        ("size", c_int ),
        ("originLayer", c_short),
        ("originIndex", c_short),
        ("targetLayer", c_short),
        ("targetIndex", c_short),
        ("style", c_ushort), # style de lien synaptique. Constantes ls...
        ("inputBool", c_ushort), 
        ("outputBool", c_ushort), # Pour AbsorberNeurone si <> 0, absorbtion des synapses enfants
        ("value", c_double),
        ("_name", c_byte* 256),
        ("netIndex", c_int),
        ("index", c_int)]
    
    def __repr__(self):
        st = super(SynapseData, self).__repr__()
        st2 = '(%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)'% (self.originLayer, 
            self.originIndex, self.targetLayer, self.targetIndex, self.style, 
            self.inputBool, self.outputBool, self.value, self.sname, self.netIndex, 
            self.index)
        return st + st2

    def __init__(self, name='', originlayer=-2, originindex=-1, targetlayer=-2, 
                 targetindex=-1, style=0, inputbool=0, outputbool=0, 
                 netindex=-1, synindex=-1, value=0.0):
        super(SynapseData, self).__init__()
        memset(byref(self), 0, sizeof(self))
        self.size = sizeof(self)
        self.originLayer = originlayer
        self.originIndex = originindex
        self.targetLayer = targetlayer
        self.targetIndex = targetindex
        self.style = style
        self.inputBool = inputbool
        self.outputBool = outputbool
        self.value = value
        self.sname = name
        self.netIndex = netindex
        self.index = synindex
    
    @Property
    def sname(self):
        """
        Property.
        
        Link sname. 
        """
        return self._getstr_at(self.__class__._name)
    @setter(sname)
    def _setname(self, value):
        self._setstr_at(self.__class__._name, value)

#---------------------------------------
class NodeData (baseData):  #  commun avec Delphi
    # !!! ATTENTION !!! Correspond au TNeuronDataEx du code Delphi
    """
    Network node data.
    
    Used for GetNeuroneInfo, or ModifyNetwork, with styles:
    
        - ModifierNeurone
        - AjoutNeurone
        - SupprimerNeurone
    
    Fields :
    
    :size: Current record size
    :_name: Node name
    :modelName: Model name
    :layerIndex: Node layer index
    :nodeIndex: Node index in its layer
    :activation: activation function
    :nodeID: Node ID
    :nodeType: Type of node
    :stylePos: Type of positioning
    :nodePos: Node position
    :activationPlus: Second activation function
    :indexNet: Model Index in multi model
    :value: Node value
    """
    #===========================================================================
    # __doc__ = Class_doc("""Données de noeud de modèle neuronal. Utilisé pour :
    # GetNeuroneInfo, ou ModifyNetwork, avec les styles:
    #    ModifierNeurone
    #    AjoutNeurone
    #    SupprimerNeurone
    # et pour les modifications de nom de modèle.""")
    #===========================================================================
    _fields_ = [
        ("size", c_int ),        
        ("_name", c_byte*256),
        ("modelName", c_int),
        ("layerIndex", c_int),
        ("nodeIndex", c_int),
        ("activation", c_int),
        ("nodeID", c_int),
        ("nodeType", c_int),     # TNeuronType : (ntNone, ntSigma, ntSigma2, ntPi, ntSigmaPlus)
        ("stylePos", c_int),       # TPositionStyle : (psBegin, psEnd)
        ("nodePos", c_int), # TAddNeuronType : (anDontCare, anConstant, anInput, anOutPut, anOutState)
        ("activationPlus", c_int),
        ("indexNet", c_int),
        ("value", c_double)]

    def __repr__(self):
        st = super(NodeData, self).__repr__()
        st2 = '(%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)'% (self.name, 
            self.modelName, self.layerIndex, self.nodeIndex, C.ACTIV_NAME[self.activation], self.nodeID, 
            self.nodeType, self.stylePos, self.nodePos, self.value, 
            C.ACTIV_NAME[self.activationPlus], self.indexNet)
        return st + st2
    
    def __init__(self, name='', isnetname=0, layerindex=-2, nodeindex=-1, 
                 activation='TANH', nodeID=-1, nodetype=1, style=0, 
                 nodeposition=0, value=0.0, netindex=-1, Model=None):
        super(NodeData, self).__init__()
        memset(byref(self), 0, sizeof(self))
        self.size = sizeof(self)
        if Model:
            self.name = Model.Name 
            self.modelName = Model.NetWorkName
            self.layerIndex = Model.LayerIndex
            self.nodeIndex = Model.NeuronIndex 
            self.activation = Model.Activation
            self.nodeID = Model.NeuronID
            self.nodeType = Model.NeuronType
            self.stylePos = Model.StylePos
            self.nodePos = Model.NeuronPosition
            self.value = Model.value
            self.activationPlus = Model.ActivationPlus
            self.indexNet = Model.IndexNet
        else:
            self.name = name 
            self.modelName = isnetname
            self.layerIndex = layerindex
            self.nodeIndex = nodeindex 
            self.activation = C.getActivIndex(activation)
            self.nodeID = nodeID
            self.nodeType = nodetype
            self.stylePos = style
            self.nodePos = nodeposition
            self.value = value
            self.activationPlus = 0
            self.indexNet = netindex
        
    @Property
    def name(self):
        """
        Property.
        
        Node name. 
        """
        return self._getstr_at(self.__class__._name)
    @setter(name)
    def _setname(self, value):
        self._setstr_at(self.__class__._name, value)

#---------------------------------------
class LayerData (baseData):  #  commun avec Delphi
    """
    Layer data.
    
    Used for ModifyNetwork with styles :
    
    - NouvelleCouche
    - DetruireCouche
    
    Fields :
    
    :size: Current record size
    :index: Layer index in the model
    :length: Layer length
    :activation: Common node activation
    :nodeType: Common type of nodes
    :stylePos: Positionning style
    :activationPlus: Secundary activation
    :_name: Base name for nodes
    :netIndex: Index of model in multi models
    :defaultnodevalue: Default nodes value
    """
#    __doc__ = Class_doc("""Données de couche. 
#    Utimisée pour  ModifyNetwork avec les styles :
#        NouvelleCouche
#        DetruireCouche.""")
    _fields_ = [    
        ("size", c_int ),
        ("index", c_short),
        ("length", c_short),
        ("activation", c_short),
        ("nodeType", c_ushort),
        ("stylePos", c_ushort),
        ("activationPlus", c_short),
        ("_name", c_byte*256),
        ("netIndex", c_int),
        ("defaultnodevalue", c_double),]

    # !!! attention changement de valeur par defaut  index=0 -> 1
    def __init__(self, name='', index=1, length=0, activfunc='TANH', ntype=1, style=0, defaultnodevalue=0.0):
        super(LayerData, self).__init__()
        memset(byref(self), 0, sizeof(self))
        self.size = sizeof(self)
        self.index = index
        self.length = length
        self.activation = C.getActivIndex(activfunc)
        self.nodeType = ntype
        self.stylePos = style
        self.activationPlus = 0
        self.name = name
        self.netIndex = 0
        self.defaultnodevalue = defaultnodevalue

    def __repr__(self):
        st = super(LayerData, self).__repr__()
        st2 = '(%s, %s, %s, %s, %s, %s, %s, %s, %s)'% (self.index, self.length, 
            C.ACTIV_NAME[self.activation], self.nodeType, self.stylePos, 
            C.ACTIV_NAME[self.activationPlus], self.name, self.netIndex,
            self.defaultnodevalue)
        return st + st2

    @Property
    def name(self):
        """
        Property.
        
        layer nodes name
        """
        return self._getstr_at(self.__class__._name)
    @setter(name)
    def _setname(self, value): 
        self._setstr_at(self.__class__._name, value)

#---------------------------------------
class MergeData( baseData ):  #  commun avec Delphi
    """
    Model merging data.
    
    Fields :
    
    :size: Current record size
    :_source: File name for the model to merge
    :inMerge: Merging of inputs
    :layerShift: Shift in layer index
    :nodeShift: Shift in node index
    :autoMerge: Merge of current model with itself
    :plugIn: Plugin the merged model if possible
    :sharedSynapses: Share the links between models
    :byName: Merge choice by node names 
    """
#    __doc__ = Class_doc("""Donnees de fusion.
#    Utilise par modifyModel avec le style 
#        MERGEMODEL.""")
    _fields_ = [("size", c_int ),
                ("_source", c_byte*256),
                ("inMerge", c_short),
                ("layerShift", c_short),
                ("nodeShift", c_short),
                ("autoMerge", c_ushort),
                ("plugIn", c_ushort),
                ("sharedSynapses", c_ushort),
                ("byName", c_ushort)]
    
    def __init__(self, source="", ninputmerge=0, nlayershift=0, nnodeshift=0,
        automerge=0, plugin=0, sharedsynapses=0, byname=0):
        super(MergeData, self).__init__()
        memset(byref(self), 0, sizeof(self))
        self.size = sizeof(self)
        self.source = source
        self.inMerge = ninputmerge
        self.layerShift = nlayershift
        self.nodeShift = nnodeshift
        self.autoMerge = automerge
        self.plugIn = plugin
        self.sharedSynapses = sharedsynapses
        self.byName = byname

    def __repr__(self):
        st = super(MergeData, self).__repr__()
        st2 = '(%s, %s, %s, %s, %s, %s, %s, %s)'% (self.source, self.inMerge, 
            self.layerShift, self.nodeShift, self.autoMerge, self.plugIn, 
            self.sharedSynapses, self.byName)
        return st + st2

    @Property
    def source(self):
        """
        Propertry.
        
        merging model source.
        """
        return self._getstr_at(self.__class__._source)
    @setter(source)
    def _setsource(self, value): 
        self._setstr_at(self.__class__._source, value)

#---------------------------------------
class ModelData( baseData ):  #  commun avec Delphi
    """
    Model data.
    
    Used by ndkInterface, createNewModel, recreateModel
    
    Fields :
    
    :size: Current record size
    :input: Number of inputs
    :output: Number of outputs
    :hidden: Number of hidden nodes
    :activation: activation function of hidden nodes
    :layer: layer of loop origin node
    :node: index in its layer if loop origin node
    :state: number of loops
    :redondancy: Take care of redundancy while creating loops
    :notCreateSynapses: Dont create synapses
    :existingNode: Origin node is an exissting node
    :polyType: Polynom type
    :_name: Model name
    """
#    __doc__ = Class_doc("""Donnees de modele - 
#    Utilise par
#        ndkInterface 
#        createNewModel
#        recreateModel.""")
    _fields_ = [( "size", c_int ),
                ( "input", c_int ),
                ( "output", c_int ),
                ( "hidden", c_int ),
                ( "activation", c_short ),
                ( "layer", c_short ),
                ( "node", c_short ),
                ( "state", c_short ),
                ( "redondancy", c_byte ),
                ( "notCreateSynapses", c_byte ),
                ( "existingNode", c_byte ),     # pour les bouclages
                ( "polyType", c_short),
                ( "_name", c_byte*256)]

    def __init__(self, input=1, output=1, hidden=2, activfunc="TANH", 
            statelayer=-1, statenodeindex=-1, order=0, nosynapse=0, 
            existingnode=0, polytype=0, name=""):
        super(ModelData, self).__init__()
        memset(byref(self), 0, sizeof(self))
        self.size = sizeof(self)
        self.input = input
        self.output = output
        self.hidden = hidden
        self.activation = C.getActivIndex(activfunc)
        self.layer = statelayer
        self.node = statenodeindex
        self.state = order
        self.redondancy = 1
        self.notCreateSynapses = nosynapse
        self.existingNode = existingnode
        self.polyType = polytype
        #if isinstance(name, str):
        self.name = str(name)            
        
    def __repr__(self):
        st = super(ModelData, self).__repr__()
        st2 = '(%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)'% (self.input, 
            self.output, self.hidden, self.activation, self.layer, self.node, 
            self.state, self.redondancy, self.notCreateSynapses, 
            self.existingNode, self.polyType, self.name)
        return st + st2

    @Property
    def name(self):
        """
        Property
        
        Model name.
        """
        return self._getstr_at(self.__class__._name)
    @setter(name)
    def _setname(self, value): 
        self._setstr_at(self.__class__._name, value)

#TModelData = ModelData # compatibilité
#---------------------------------------
class DataDrivenModel (baseData): # commun avec delphi TDataDrivenModel
    """
    Data for data driven model creation.
    
    Fields :
    
    :size: Current record size
    :input: Number of inputs
    :output: Number of outputs
    :hidden0: Minimum number of hidden nodes
    :hiddenM: Maximum number of hidden nodes
    :order: Order of model
    :polyType: Type of polynom
    :normalize: normalisation action
    :_savedir: saving folder
    :_tempdir: temporary folder
    :dataFormat: format of data 
    """
#    __doc__ = Class_doc("""Donnees de creation d'un modele pilote par les donnees""")
    _fields_ = [
        ("size", c_long),
        ("input", c_long),
        ("output", c_long),
        ("hidden0", c_long),
        ("hiddenM", c_long),
        ("order", c_long),
        ("polyType", c_long),
        ("_reserve1", c_long),
        ("normalize", c_long),
        ("_reserve2", c_long),
        ("_savedir", c_byte*256),
        ("_tempdir", c_byte*256),
        ("dataFormat", c_long*256)]
    
    def __init__(self, inputs=-1, outputs=-1, hidden0=-2, hiddenm=5, order=0, 
                 polyt=0, normalize=1, savedir="", tempdir="", dataformat=[]):
        super(DataDrivenModel, self).__init__()
        memset(byref(self), 0, sizeof(self))
        self.size = sizeof(self)
        self.input = inputs
        self.output = outputs
        self.hidden0 = hidden0
        self.hiddenM = hiddenm
        self.order = order
        self.polyType = polyt
        self.normalize = normalize
        self.savedir = savedir
        self.tempdir = tempdir
        for ind in range(256):
            self.dataFormat[ind] = -1
        for ind, item in enumerate(dataformat):
            self.dataFormat[ind] = int(item)

    def __repr__(self):
        st = super(DataDrivenModel, self).__repr__()
        st2 = '(%s, %s, %s, %s, %s, %s, %s, %s, %s, %s)'% (self.input, self.output, 
            self.hidden0, self.hiddenM, self.order, self.polyType, self.normalize,
            self.saveDir, self.tempDir, self.dataFormat)
        return st + st2
    
    @Property
    def savedir(self):
        """
        Property.
        
        Saving folder.
        """
        return self._getstr_at(self.__class__._savedir)
    @setter(savedir)
    def _setsavedir(self, value):
        self._setstr_at(self.__class__._savedir, value)

    @Property
    def tempdir(self):
        """
        Property.
        
        Temporary folder.
        """
        return self._getstr_at(self.__class__._tempdir)
    @setter(tempdir)
    def _settempdir(self, value): 
        self._setstr_at(self.__class__._tempdir, value)

#-------------------------------------------------------------------------------
class DataDrivenModelTrain (baseData):
    """
    Training data for data driven model.
    
    Fields :
    
    :size: Current record size
    :inits: Number of parameters initialisation per model
    :style: Training style
    :criterion: Choice criterion
    :algorithm: Training algorithm
    :epochs: Number of epochs for each parameters initialisation
    :_subdir: Saving sub-folder
    :_target: Target file name
    :onFlightCriterion: On flight selection criterion
    """
    #__doc__ = Class_doc("""donnees d'apprentissage d'un model pilote par les donnees""")
    _fields_ = [
        ("size", c_long),
        ("inits", c_long),
        ("style", c_long),
        ("criterion", c_long),
        ("algorithm", c_long),
        ("epochs", c_long),
        ("_subdir", c_byte*256),
        ("_target", c_byte*256),
        ("onFlightCriterion", c_long)]
    
    def __init__(self, inits=10, style=C.TR_S_INIT_PARAM_LEVERAGE, 
                 onFlightCriterion=C.TCR_STD_DEV_BIAS_LESS, 
                 criterion=C.TCR_STD_DEV_BIAS_LESS,
                 algorithm=C.TA_LM, epochs=50, subdir="", target=""):
        super(DataDrivenModelTrain, self).__init__()
        memset(byref(self), 0, sizeof(self))
        self.size = sizeof(self)
        self.inits = inits
        self.style = style
        self.criterion = criterion
        self.algorithm = algorithm
        self.epochs = epochs
        self.subDir = subdir
        self.target = target
        self.onFlightCriterion = onFlightCriterion 

    def __repr__(self):
        st = super(DataDrivenModelTrain, self).__repr__()
        st2 = '(%s, %s, %s, %s, %s, %s, %s, %s)'% (self.inits, self.style, 
            self.criterion, self.algorithm, self.epochs, self.subDir, self.target, 
            self.onFlightCriterion)
        return st + st2
    
    @Property
    def subdir(self):
        """
        Property.
        
        Saving subdir.
        """
        return self._getstr_at(self.__class__._subdir)
    @setter(subdir)
    def _setsubdir(self, value): 
        self._setstr_at(self.__class__._subdir, value)

    @Property
    def target(self):
        """
        Property.
        
        Target folder
        """
        return self._getstr_at(self.__class__._target)
    @setter(target)
    def _settarget(self, value):
        self._setstr_at(self.__class__._target, value)

#-------------------------------------------------------------------------------
if __name__ == "__main__":
    print(baseData.__doc__)
    print()
    print(LoopData.__doc__)
    print()
    print(SynapseData.__doc__)
    print()
    print(NodeData.__doc__)
    print()
    print(LayerData.__doc__)
    print()
    print(MergeData.__doc__)
    print()
    print(ModelData.__doc__)
    print()
    print(DataDrivenModel.__doc__)
    print()
    print(DataDrivenModelTrain.__doc__)
    


#-*- coding: ISO-8859-15 -*-
#===============================================================================
# $Id: _drivermultidyn.py 4787 2018-05-27 20:46:03Z jeanluc $
#  Module  _drivermultidyn
#  Projet MonalPy
# 
#  Implementation python de monal
# interface pour Dll
#
#  Author: Jean-Luc PLOIX  -  NETRAL
#  Mai 2014
#
# Copyright 2016 Jean-Luc PLOIX
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import os, sys
from ctypes import CFUNCTYPE, byref, c_int, c_long, c_double, c_void_p
from weakref import proxy
from numpy import asarray, ndarray, zeros, load, NaN, reshape, sqrt
from numpy import int as npint
from numpy import double as npdouble
from numpy import float as npsum
from numpy import sum as npflaot
from numpy import dot as npdot

from metaphor.nntoolbox.utils import floatEx, setRandSeed
from metaphor.monal.Property import Property #, Properties, delegate
from metaphor.monal import monalconst as C, specialmath as sp
from metaphor.monal.lcode import codeconst as CC
from metaphor.nntoolbox.utils import maxWorkers
from metaphor.monal.driver import DriverBase  #, getRandState, setRandState
from metaphor.monal.driver._driverlib import DriverLibError, DriverLib
from metaphor.monal.driver import optim_bfgs, optim_leastsq

import math
# try:
#     from sc ipy.stats.t import interval
#     from sc ipy import optimize
#     OKSc ipy = True
# except ImportError:
#     OKSc ipy = False
#     from monal.util.monaltoolbox import interval
#import numpy.ran dom as rnd
# import concurrent.fut ures as cf

# U SE_PARALLEL_MAIN = 1
# U SE_SUPER_TRAIN = 1
# D EBUG_LOO = 0
# D EBUG_LOO_FILE = 0
# D  EBUG_LOCAL_TRAIN = 0


class DriverMultiDynError (DriverLibError): 
    """DriverMultiDyn Exception.
    """
    pass

class DriverMultiDyn(DriverLib):
    """class DriverMultiDyn
    Gestionnaire des drivers d'apprentissage multimodeles provenant de 
    librairies dynamiques (dll, so, dylib);
    """
    def __init__(self, libfilename, verbose=0, callback=None): # DriverMultiDyn
        super(DriverMultiDyn, self).__init__(libfilename, verbose)  #, callback
        #self._modelType = 2
        #self.model Count = self.dataCount
        if callback:
            self.setcallback(callback)
        
    def __repr__(self):
        return self.repr('long')
    
    @property
    def modelCount(self):
        return self.dataCount
    
    @Property
    def configuration(self): # DriverMultiDyn
        return self.getStrPropertyFromLib("configuration")
    
    @Property
    def configurationLong(self): # DriverMultiDyn
        return self.getStrPropertyFromLib("configuration")
#         return self.getStrPropertyFromLib("configurationlong")
    
    def reprlist(self, mode='long', prefix="module"):
        lst = super(DriverMultiDyn, self).reprlist(mode, prefix)
        lst.append("model count: %s"%  self.modelCount)
        if mode in ['long']:  #, 'short'
            lst.append("configuration: %s"% self.configuration)
            lst.append("signature: {0}".format(self.signature()))
        return lst    
    
    def dict(self):
        return {
            'name': self.moduleName,
            'type': self.moduleType,
            'codeversion': self.codeVersion,
            'created': self.created,
            'mark': self.mark,
            'base': self.base,
            'config': self.configuration,
            'property': self.propertyName,
            'paramCount': self.paramCount,
            'modelCount': self.modelCount,
            'hidden': self.hidden,
            }
    
    def __getstate__(self):
        objdict = self.__dict__.copy()
        objdict["lib"] = None
        objdict["libname"] = ""
        objdict["_lastweights"] = None
        objdict["_callback"] = None
        objdict["targets"] = self.targets
        objdict["propertyName"] = self.propertyName
        return objdict
     
    def __setstate__(self, state):
        targets = state.pop("targets", None)
        propertyName = state.pop("propertyName", None)
        self.__dict__.update(state)
        self._setlib(self.libfilename)
        if targets is not None:
            self.targets = targets
        if propertyName is not None:
            self.propertyName = propertyName
         
    def signature(self):
        weights = [0.01 for _ in range(self.paramCount)]
        inputs = [0.01 for _ in range(self.inputCount)]
        res = 0
        for ind in range(self.modelCount):
            if self.inputCount:
                res += self.transferindex(weights, inputs, ind)
            else:
                res += self.transferindex(weights, ind)
        return res / self.paramCount
    
    def transferindex(self, params=None, index=-1):
        if isinstance(params, str) and params == 'extra':
            return [self.transferindex(param, index) for param in self._extraweights]
        if params is not None:
            params = asarray(params, npdouble)
        else:
            if self.weights is not None:
                params = self.weights
        if params is not None:
            self._lastweights = params.copy
        else:
            raise DriverMultiDynError("params sould not be null in transferindex")
         
        out = None
        try:
            source = params.copy()
            source_p = c_void_p(source.ctypes.data)
            out = zeros((1,), dtype=npdouble)
            out_p = c_void_p(out.ctypes.data)
            res = self.lib.transferindex(index, source_p, out_p)
            if res < 0: 
                raise DriverMultiDynError(self.lasterror)
             
        except: pass
        if self.outputCount ==1:
            out = out[0]
        return out
 
    def transfergradientindex(self, params=None, index=-1):
        if params is not None:
            params = asarray(params, npdouble)
        else:
            if self.weights is not None:
                params = self.weights
        if params is not None:
            self._lastweights = params.copy
        else:
            raise DriverMultiDynError("params sould not be null in transfergradientindex")
         
        out = None
        #target = None
        try:
            paramCount = self.lib.paramCount()
            source = params.copy()  #asarray(params, npdouble)
            source_p = c_void_p(source.ctypes.data)
            gradient = zeros((paramCount,), dtype=npdouble)
            gradient_p = c_void_p(gradient.ctypes.data)
            out = zeros((1,), dtype=npdouble)
            out_p = c_void_p(out.ctypes.data)
            res = self.lib.transfergradientindex(index, source_p, out_p, gradient_p)
            if res < 0: 
                raise DriverMultiDynError(self.lasterror)
             
        except: pass
        if self.outputCount ==1:
            out = out[0]
        return out, gradient
         
def onGetNewWeights(model, weightsstddev):
    return model.newWeights(weightsstddev)
    
def kerneltrainintegratedmain(model, weights, maxiter, withhistory, index=-1):
    res = model.kerneltrainintegratedmain(weights, maxiter, withhistory, index)
    if index < 0:
        return res
    return (index,) + res

def core_train2(model, weights, maxiter, withhistory, index=-1):
    return model.core_train2(weights, maxiter, withhistory, index)

def kerneltrainintegratedmain3(model, weights, maxiter, withhistory, index=-1):
    return model.kerneltrainintegratedmain3(weights, maxiter, withhistory, index)

#==============================================================================
if __name__ == "__main__":
    from io import StringIO, BytesIO
    import pickle as pickle
    single = 1
#     source = "C:\\Projets\\GM\\TestAD\\Base115new\\base115new_chi0iso00n.dll"
#     source = "/Users/Shared/GMWorkspace/BaseA109/libbasea109_chi03n.so"
    source = "/Users/jeanluc/Documents/LiClipseWorkspace/Metaphor_Test/src/metaphor/test_monal/testFiles/libnn_5_5_1.so"
    #"/Users/Shared/workspace/GraphMachine/test/testfiles/libbase321_27_dgrs5_5n.so"
    targetfile = "/Users/jeanluc/Documents/LiClipseWorkspace/Metaphor_Test/src/metaphor/test_monal/testFiles/L_153_target.csv" 
    #"/Users/Shared/workspace/GraphMachine/test/testfiles/LogP_Base321"

#code version
    
    model = DriverMultiDyn(source)
    #model.settrainingset()
    model.setRandSeed(1947)
    model.targets = targetfile
    
#    st = pickle.dumps(model)
    stream = BytesIO()
    pickle.dump(model, stream)
    stream.flush()
    stream.seek(0)
    #stream2 = StringIO(stream.getvalue())
    model2 = pickle.load(stream, allow_pickle=True)
    stream.close()
    #model2 = DriverMultiDyn(source)
    #model.settrainingset()
    #model.set RandSeed(1947)
    #model2.targets = targetfile
    
    print("property:", model.propertyName)
    print()
    print("moduleName", model.moduleName)
    print("monalVersion", model.monalVersion)
    print("created", model.created)
    print("mark", model.mark)
    print("base", model.base)
    print("linear", model.isLinear)
    print("paramlCount", model.paramCount)
    print("modelCount", model.modelCount)
    print("inputCount", model.inputCount)
    print("outputCount", model.outputCount)
    print("hidden", model.hidden)
    print("biasBased", model.biasbased)
    print("outlinks", model.outlinks)
    #model2.targets = [0 for _ in range(model.modelCount)]
    #print "target 1", model.targets
    #print "target 2", model2.targets
    print("norm 1", model.outputnorm)
    print("norm 2", model2.outputnorm)
    print(model.lib)
    print(model2.lib)
    
    #if single:
    model.setRandSeed(194999)
    ww = model.newWeights(0.3)
    #residu = model.getres iduals(ww)
    #print "residu", residu
    costini = model.getCost(ww)
    #ww2, epochs, cost, code
    #bests, saveres = model.mult itrain(initcount=5, initweights=ww, maxiter=100)
    #bestmem = bests[0][0]
    #print "bests"
    #for val in bests:
    #    print "", val
    
    
    
    
 #    cur, epochs, cost, trai nend, code = model.kerneltrainintegratedmain(ww, 150)
    #cost2 = model.getcost(cur)
    #press_old = model.getPRESS_(cur, None, True, True)
#     press = model.getPRESS(cur, 1, 1)
#     print("cost ini  ", costini)
#     print("cost final", cost)
#     print("PRESS", press)
#     #print "PRESS old", press_old
#     print("epochs", epochs)
#     #else:
        #os.system("ipcluster start")
        #lst = multitrainintegrated(model, targetfile, None, 8, 100)
        #for val in lst:
        #    print val
        #for ind, (cur, N, fopt, mem, code) in enumerate(lst):
        #    print ind, N, fopt
        
        #os.system("ipcluster stop")
    
    #model = None
    
    print("done")


# -*- coding: ISO-8859-1 -*-
#-------------------------------------------------------------------------------
# $Id$
#
# Copyright 2016 Jean-Luc PLOIX
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#-------------------------------------------------------------------------------
'''
Created on 2 mars 2017

@author: jeanluc
'''

import os, sys, re, stat
from site import getuserbase, USER_BASE
from distutils.dir_util import mkpath
from distutils import sysconfig as sc
import distutils.ccompiler as dc
from distutils import errors as de
import shutil, errno
from tempfile import mkdtemp

# Fonctions utilitaires pur le setup.py
#----------------------------------------------------------------------------

def hasPackage(pth, package="monal", pythonver="3.6", version=""):  #, versionmin=""):
    content = os.listdir(pth)
    res = package in content
    if not res:
        return False
    if not len(version):
        return True
#     if len(pythonver) and ('.' in pythonver):
#         egg_info = "%s-%s-py%s.egg-info" % (package, version, pythonver)
#         if len(version):
#             if (package in content) and (egg_info in content):
#                 return True
#     else:
#     egg_info = "%s-%s-py%s" % (package, version, pythonver)
    egg_info = "%s-%s" % (package, version)
    for val in content:
        if val.startswith(egg_info):
            return True
    return False

# def compile monal lib(erasesources=True, userTarget=False, outfile="", monalpath="", isVirtual=False, isWin=False):
#     #import distutils.ccompiler as dc
#     #import distutils.sysconfig as sc
#     
#     libname = 'monal'
#     modulename = 'monal'
#     #version = __import__('monal').__version__
#     #applidir = getApplicationsDir(libname)
#     sitebasepath = getsitebase(userTarget, isVirtual)
#     #egg_info_path = os.path.join(sitebasepath, "%s-%s.egg-info"% (modulename, version))
#     #egg_info_sourcepath = os.path.join(egg_info_path, "SOURCES.txt")
#     if not len(monalpath):
#         monalpath = os.path.join(sitebasepath, modulename)
#         if not os.path.exists(monalpath):
#             monalpath = os.path.join(os.path.dirname(os.path.abspath(__file__)), modulename)
#     comp = dc.get_default_compiler()
#     source_dir = os.path.join(monalpath, 'lcode', 'code_templates', 'common')
#     target_dir = os.path.join(monalpath, 'lcode')
#     try:
#         curmode = os.stat(target_dir).st_mode
#         os.chmod(target_dir, curmode | stat.S_IWOTH | stat.S_IROTH) 
#         OKmode = True
#     except:
#         OKmode = False
#     include_dirs = [sc.get_config_vars()['INCLUDEPY'], os.path.join(monalpath, 'include')]
#     excluded = ['readfile', 'dlmdriver']
#     extra_args = ['-fPIC']  
#     if outfile:
#         outfile.write( "libname : %s\n"% libname)
#         #outfile.write( "applidir : %s\n"% applidir)
#         outfile.write( "monalpath : %s\n"% monalpath)
#         outfile.write( "source_dir : %s\n"% source_dir)
#         outfile.write( "target_dir : %s\n"% target_dir)
#         outfile.write( "include_dirs :\n") 
#         for val in include_dirs:
#             outfile.write("\t%s\n"% val)
# 
#     try:
#         rr = create_lib(comp, libname, source_dir=source_dir, 
#             target_dir=target_dir, include_dirs=include_dirs, excluded=excluded,
#             extra_preargs=extra_args, isWin=isWin) 
#         if erasesources:
#             eraseCommon(userTarget, outfile, monalpath, isVirtual)
#     except Exception as err:
#         rr = -1
#         if outfile:
#             outfile.write( "Error in compiling monal lib\n")
#             outfile.write(str(err))
#             outfile.write("\n")
#         else:
#             print("Error in compiling monal lib\n")
#     if OKmode:
#         os.chmod(target_dir, curmode)
#     return rr

# def eraseCommon(userTarget=False, outfile="", monalpath="", isVirtual=False):
#     from distutils.dir_util import  remove_tree
#     if not len(monalpath):
#         libname = 'monal'
#         monalpath = os.path.join(getsitebase(userTarget, isVirtual), libname)
#     source_dir = os.path.join(monalpath, 'lcode', 'code_templates', 'common')
#     remove_tree(source_dir, verbose=True)
# #     if outfile:
# #         outfile.write( "removing %s\n"% source_dir)

def getChildPath(parent, folder="site-packages"):
    """Look for a 'folder' directory in the children tree of 'parent'.
    return empty string if not found.
    """
    res = ""
    for path, dirs, _ in os.walk(parent):
        if folder in dirs:
            return os.path.join (path, folder)
        else:
            for direc in dirs:
                res = getChildPath(direc, folder)
                if res: break  #return res
    return res
 
def getbase(userTarget=False, isVirtual=False):
    if isVirtual or not userTarget:
        return os.path.abspath(sys.prefix)
    if userTarget:
        try:
            return getuserbase()
        except:
            return USER_BASE
    
def getsitebase(targetuser=False, isVirtual=False):
    pth = getbase(targetuser, isVirtual)
    return getChildPath(pth, "site-packages")

def getApplicationsDir(appli="", docreate=False):
    base = os.path.join(os.path.expanduser("~"), "Applications")
    if appli:
        base = os.path.join(base, appli)
    if docreate:
        mkpath(base)
    return base
    
def compilerName():
    comp = dc.get_default_compiler()
    getnext = False

    for a in sys.argv[2:]:
        if getnext:
            comp = a
            getnext = False
            continue
    #separated by space
        if a == '--compiler'  or  re.search('^-[a-z]*c$', a):
            getnext = True
        continue
    #without space
        m = re.search('^--compiler=(.+)', a)
        if m == None:
            m = re.search('^-[a-z]*c(.+)', a)
        if m:
            comp = m.group(1)

    return comp

def create_lib(comp, libname, source_dir="", target_dir="", include_dirs=[], 
               excluded=[], extra_preargs=[], doprint=False, isWin=False):
    """Creation of a static library.
    parameters :
        comp -> compiler name. Must be in the list of available compilers.
        libname -> name of the library. Not file name.
        source_dir -> directory of C sources files.
        target_dir-> directory of resulted library file. 
            If empty, library will be installed in Python libs folder. 
        excluded -> list of excluded files. '.c' extension will be added if necessary.
        doprint -> print results.
    """
    res = 0 
    cur = os.getcwd()
    for i, val in enumerate(excluded):
        if not val.lower().endswith(".c"):
            excluded[i] = val + ".c"
    if not target_dir:
        target_dir = os.path.join(sys.prefix, "libs")
    os.chdir(source_dir)
    if not len(include_dirs):
        incpy = sc.get_config_vars()['INCLUDEPY']
        include_dirs = [incpy, os.path.join(source_dir, 'include')]
    comp = comp if comp else compilerName()
    cc = dc.new_compiler(compiler=comp)
    cc.mkpath(target_dir)
    _, _, files = next(os.walk(source_dir))
    #sources = [os.path.join(root, val) for val in files if val.endswith(".c") and not val in excluded]
    sources = [val for val in files if val.endswith(".c") and not val in excluded]
    macros = [("_CRT_SECURE_NO_WARNINGS", 1)]
    #plat = sys.platform
    #if (plat=='nt') or plat.startswith('win'):
    if isWin:  #IS_WIN32 or IS_WINDOWS:
        macros.append(("WIN", 1))
    try:
        temp = mkdtemp(prefix="$%s"% libname)  
        fulllibname = os.path.join(target_dir, 'lib%s.a'% libname)
        if os.path.exists(fulllibname):
            os.remove(fulllibname)
        objlist = cc.compile(sources, output_dir=temp, macros=macros, 
            include_dirs=include_dirs, extra_preargs=extra_preargs)
        cc.create_static_lib(objlist, libname, target_dir)
        res = 1
        if doprint:
            print(("compiler %s succeded"% comp))
    except de.CompileError:
        print(("compiler %s failed"% comp))
    finally:
        try:
            shutil.rmtree(temp)  # delete directory
        except OSError as exc:
            if exc.errno != errno.ENOENT:  # ENOENT - no such file or directory
                raise  # re-raise exception
    os.chdir(cur)
    return res
#----------------------------------------------------------------------------

if __name__ == '__main__':
    """Recreation of the monal module
    """
    monalpath = os.path.dirname(__file__)
    monalpath = os.path.dirname(monalpath)
#    compile monallib(erasesources=False, userTarget=False, outfile="", monalpath=monalpath)
    print('Done')
